# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import session
from oslo_log import log as logging

from congress import exception as congress_exc

LOG = logging.getLogger(__name__)

_FACADE = None


def _create_facade_lazily():
    global _FACADE

    if _FACADE is None:
        _FACADE = session.EngineFacade.from_config(cfg.CONF, sqlite_fk=True)

    return _FACADE


def get_engine():
    """Helper method to grab engine."""
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(autocommit=True, expire_on_commit=False, make_new=False):
    """Helper method to grab session."""
    if make_new:  # do not reuse existing facade
        facade = session.EngineFacade.from_config(cfg.CONF, sqlite_fk=True)
    else:
        facade = _create_facade_lazily()
    return facade.get_session(autocommit=autocommit,
                              expire_on_commit=expire_on_commit)


def get_locking_session():
    """Obtain db_session that works with table locking

    supported backends: MySQL and PostgreSQL
    return default session if backend not supported (eg. sqlite)
    """
    if is_mysql() or is_postgres():
        db_session = get_session(
            autocommit=False,
            # to prevent implicit new transactions,
            # which UNLOCKS in MySQL
            expire_on_commit=False,  # need to UNLOCK after commit
            make_new=True)  # brand new facade avoids interference

    else:  # unsupported backend for locking (eg sqlite), return default
        db_session = get_session()

    return db_session


def lock_tables(session, tables):
    """Write-lock tables for supported backends: MySQL and PostgreSQL"""
    session.begin(subtransactions=True)
    if is_mysql():  # Explicitly LOCK TABLES for MySQL
        session.execute('SET autocommit=0')
        session.execute('LOCK TABLES {}'.format(
            ','.join([table + ' WRITE' for table in tables])))
    elif is_postgres():  # Explicitly LOCK TABLE for Postgres
        session.execute('BEGIN TRANSACTION')
        for table in tables:
            session.execute('LOCK TABLE {} IN EXCLUSIVE MODE'.format(table))


def commit_unlock_tables(session):
    """Commit and unlock tables for supported backends: MySQL and PostgreSQL"""
    try:
        session.execute('COMMIT')  # execute COMMIT on DB backend
        session.commit()
        # because sqlalchemy session does not guarantee
        # exact boundary correspondence to DB backend transactions
        # We must guarantee DB commits transaction before UNLOCK

        # unlock
        if is_mysql():
            session.execute('UNLOCK TABLES')
        # postgres automatically releases lock at transaction end
    except db_exc.DBDataError as exc:
        LOG.exception('Database backend experienced data error.')
        raise congress_exc.DatabaseDataError(data=exc)


def rollback_unlock_tables(session):
    """Rollback and unlock tables

    supported backends: MySQL and PostgreSQL
    """
    # unlock
    if is_mysql():
        session.execute('UNLOCK TABLES')

    # postgres automatically releases lock at transaction end

    session.rollback()


def is_mysql():
    """Return true if and only if database backend is mysql"""
    return (cfg.CONF.database.connection is not None and
            (cfg.CONF.database.connection.split(':/')[0] == 'mysql' or
             cfg.CONF.database.connection.split('+')[0] == 'mysql'))


def is_postgres():
    """Return true if and only if database backend is postgres"""
    return (cfg.CONF.database.connection is not None and
            (cfg.CONF.database.connection.split(':/')[0] == 'postgresql' or
             cfg.CONF.database.connection.split('+')[0] == 'postgresql'))


def is_sqlite():
    """Return true if and only if database backend is sqlite"""
    return (cfg.CONF.database.connection is not None and
            (cfg.CONF.database.connection.split(':/')[0] == 'sqlite' or
             cfg.CONF.database.connection.split('+')[0] == 'sqlite'))



# Copyright (c) 2018 Orange. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Conflict with flake8
# pylint: disable = bad-continuation

"""A theory that contains rules that must be treated by Z3."""

import time

import logging
import six

from congress.datalog import base
from congress.datalog import compile as ast
from congress.datalog import nonrecursive
from congress.datalog import ruleset
from congress.datalog import unify
from congress import exception
from congress.z3 import typechecker
from congress.z3 import z3types

# pylint: disable = ungrouped-imports
MYPY = False
if MYPY:
    # pylint: disable = unused-import
    from congress.datalog import topdown                        # noqa
    from typing import Dict, Callable, Union, List, Any, Tuple  # noqa
    import z3                                                   # noqa
    Z3_RESULT = Tuple[Union[bool, List[List[z3.ExprRef]]],
                      List[ast.Variable],
                      List[z3types.Z3Type]]

LOG = logging.getLogger(__name__)

Z3_ENGINE_OPTIONS = {'engine': 'datalog'}

Z3OPT = z3types.z3

INTER_COMPILE_DELAY = 60.0


def cycle_not_contained_in_z3(theories, cycles):
    # type: (Dict[str, base.Theory], List[List[str]]) -> bool
    """Check that there is a true cycle not through Z3 theory

    A cycle is irreducible if it contains at least one element which is not a
    Z3Theory for which recursion is allowed. Cycles are represented by lists of
    qualified table names.
    """
    acceptables = [
        th.name
        for th in six.itervalues(theories)
        if isinstance(th, Z3Theory)]
    return any(fullname[:fullname.index(':')] not in acceptables
               for cycle in cycles for fullname in cycle)


# TODO(pcregut): Object constants should evolve to use the type system
# rather than custom types.
def congress_constant(val):
    """Creates an object constant from a value using its type"""
    if isinstance(val, six.string_types):
        typ = ast.ObjectConstant.STRING
    elif isinstance(val, int):
        typ = ast.ObjectConstant.INTEGER
    elif isinstance(val, float):
        typ = ast.ObjectConstant.FLOAT
    else:
        val = str(val)
        typ = ast.ObjectConstant.STRING
    return ast.ObjectConstant(val, typ)


def retrieve(theory, tablename):
    # type: (topdown.TopDownTheory, str) -> List[ast.Literal]
    """Retrieves all the values of an external table.

    Performs a select on the theory with a query computed from the schema
    of the table.
    """
    arity = theory.schema.arity(tablename)
    table = ast.Tablename(tablename, theory.name)
    args = [ast.Variable('X' + str(i)) for i in range(arity)]
    query = ast.Literal(table, args)
    return theory.select(query)


class Z3Theory(nonrecursive.RuleHandlingMixin, base.Theory):
    """Theory for Z3 engine

    Z3Theory is a datalog theory interpreted by the Z3 engine instead of
    the usual congress internal engine.
    """

    def __init__(self, name=None, abbr=None,
                 schema=None, theories=None, desc=None, owner=None):
        super(Z3Theory, self).__init__(
            name=name, abbr=abbr, theories=theories,
            schema=ast.Schema() if schema is None else schema,
            desc=desc, owner=owner)
        LOG.info('z3theory: create %s', name)
        self.kind = base.Z3_POLICY_TYPE
        self.rules = ruleset.RuleSet()
        self.dirty = False
        self.z3context = None
        Z3Context.get_context().register(self)

    def select(self, query, find_all=True):
        """Performs a query"""
        return self.z3context.select(self, query, find_all)

    def arity(self, tablename, modal=None):
        """Arity of a table"""
        return self.schema.arity(tablename)

    def drop(self):
        """To call when the theory is forgotten"""
        self.z3context.drop(self)

    def _top_down_eval(self,
                       context,  # type: topdown.TopDownTheory.TopDownContext
                       caller    # type: topdown.TopDownTheory.TopDownCaller
                       ):
        # type: (...) -> bool
        """Evaluation entry point for the non recursive engine

        We must compute unifiers and clear off as soon as we can
        giving back control to the theory context.
        Returns true if we only need one binding and it has been found,
        false otherwise.
        """
        raw_lit = context.literals[context.literal_index]
        query_lit = raw_lit.plug(context.binding)
        answers, bvars, translators = self.z3context.eval(self, query_lit)
        if isinstance(answers, bool):
            if answers:
                return (context.theory._top_down_finish(context, caller)
                        and not caller.find_all)
            return False
        for answer in answers:
            changes = []
            for (val, var, trans) in six.moves.zip(answer, bvars, translators):
                chg = context.binding.add(var, trans.to_os(val), None)
                changes.append(chg)
            context.theory._top_down_finish(context, caller)
            unify.undo_all(changes)
            if not caller.find_all:
                return True
        return False


class Z3Context(object):
    """An instance of Z3 defined first by its execution context"""

    _singleton = None

    def __init__(self):
        self.context = Z3OPT.Fixedpoint()
        self.context.set(**Z3_ENGINE_OPTIONS)
        self.z3theories = {}        # type: Dict[str, Z3Theory]
        self.relations = {}         # type: Dict[str, z3.Function]
        # back pointer on all theories extracted from registered theory.
        self.theories = None        # type: Dict[str, topdown.TopDownTheory]
        self.externals = set()      # type: Set[Tuple[str, str]]
        self.type_registry = z3types.TypeRegistry()
        self.last_compiled = 0

    def register(self, theory):
        # type: (Z3Theory) -> None
        """Registers a Z3 theory in the context"""
        if self.theories is None:
            self.theories = theory.theories
        theory.z3context = self
        self.z3theories[theory.name] = theory

    def drop(self, theory):
        # type: (Z3Theory) -> None
        """Unregister a Z3 theory from the context"""
        del self.z3theories[theory.name]

    @staticmethod
    def get_context():
        # type: () -> Z3Context
        """Gives back the unique instance of this class.

        Users should not use the class constructor but this method.
        """
        if Z3Context._singleton is None:
            Z3Context._singleton = Z3Context()
        return Z3Context._singleton

    def eval(self,
             theory,   # type: Z3Theory
             query     # type: ast.Literal
             ):
        # type: (...) -> Z3_RESULT
        """Solves a query and gives back a raw result

        Result is in Z3 ast format with a translator
        """
        theories_changed = any(t.dirty for t in self.z3theories.values())
        # TODO(pcregut): replace either with an option or find something
        # better for the refresh of datasources.
        needs_refresh = time.time() - self.last_compiled > INTER_COMPILE_DELAY
        if theories_changed or needs_refresh:
            # There is no reset on Z3 context. Replace with a new one.
            self.context = Z3OPT.Fixedpoint()
            self.context.set(**Z3_ENGINE_OPTIONS)
            self.typecheck()
            self.compile_all()
            self.synchronize_external()
        z3query = self.compile_query(theory, query)
        self.context.query(z3query)
        z3answer = self.context.get_answer()
        answer = z3types.z3_to_array(z3answer)
        typ_args = theory.schema.types(query.table.table)
        variables = []
        translators = []
        for arg, typ_arg in six.moves.zip(query.arguments, typ_args):
            if isinstance(arg, ast.Variable) and arg not in variables:
                translators.append(
                    self.type_registry.get_translator(str(typ_arg.type)))
                variables.append(arg)
        return (answer, variables, translators)

    def select(self, theory, query, find_all):
        # type: (Z3Theory, ast.Literal, bool) -> List[ast.Literal]
        """Query a theory"""
        (answer, variables, trans) = self.eval(theory, query)
        pattern = [
            variables.index(arg) if isinstance(arg, ast.Variable) else arg
            for arg in query.arguments]

        def plug(row):
            """Plugs in found values in query litteral"""
            args = [
                (congress_constant(trans[arg].to_os(row[arg]))
                 if isinstance(arg, int) else arg)
                for arg in pattern]
            return ast.Literal(query.table, args)

        if isinstance(answer, bool):
            return [query] if answer else []
        if find_all:
            result = [plug(row) for row in answer]
        else:
            result = [plug(answer[0])]
        return result

    def declare_table(self, theory, tablename):
        """Declares a new table in Z3 context"""
        fullname = theory.name + ':' + tablename
        if fullname in self.relations:
            return
        typ_args = theory.schema.types(tablename)
        param_types = [
            self.type_registry.get_type(str(tArg.type))
            for tArg in typ_args]
        param_types.append(Z3OPT.BoolSort())
        relation = Z3OPT.Function(fullname, *param_types)
        self.context.register_relation(relation)
        self.relations[fullname] = relation

    def declare_tables(self):
        """Declares all tables defined in Z3 context"""
        for theory in six.itervalues(self.z3theories):
            for tablename in theory.schema.map.keys():
                self.declare_table(theory, tablename)

    def declare_external_tables(self):
        """Declares tables from other theories used in Z3 context"""
        def declare_for_lit(lit):
            """Declares the table of a litteral if necessary"""
            service = lit.table.service
            table = lit.table.table
            if service is not None and service not in self.z3theories:
                self.externals.add((service, table))
        for theory in six.itervalues(self.z3theories):
            for rules in six.itervalues(theory.rules.rules):
                for rule in rules:
                    for lit in rule.body:
                        declare_for_lit(lit)
        for (service, table) in self.externals:
            self.declare_table(self.theories[service], table)

    def compile_facts(self, theory):
        # type: (Z3Theory) -> None
        """Compiles the facts of a theory in Z3 context"""
        for tname, facts in six.iteritems(theory.rules.facts):
            translators = [
                self.type_registry.get_translator(str(arg_type.type))
                for arg_type in theory.schema.types(tname)]
            fullname = theory.name + ':' + tname
            z3func = self.relations[fullname]
            for fact in facts:
                z3args = (tr.to_z3(v, strict=True)
                          for (v, tr) in six.moves.zip(fact, translators))
                z3fact = z3func(*z3args)
                self.context.fact(z3fact)

    def compile_atoms(self,
                      theory,    # type: Z3Theory
                      head,      # type: ast.Literal
                      body       # type: List[ast.Literal]
                      ):
        # type: (...) -> Tuple[z3.Const, z3.ExprRef, List[z3.ExprRef]]
        """Compile a list of atoms belonging to a single variable scope

        As it is used mainly for rules, the head is distinguished.
        """
        variables = {}  # type: Dict[str, z3.Const]
        z3vars = []

        def compile_expr(expr, translator):
            """Compiles an expression to Z3"""
            if isinstance(expr, ast.Variable):
                name = expr.name
                if name in variables:
                    return variables[name]
                var = Z3OPT.Const(name, translator.type())
                variables[name] = var
                z3vars.append(var)
                return var
            elif isinstance(expr, ast.ObjectConstant):
                return translator.to_z3(expr.name)
            else:
                raise exception.PolicyException(
                    "Expr {} not handled by Z3".format(expr))

        def compile_atom(literal):
            """Compiles an atom in Z3"""
            name = literal.table.table
            svc = literal.table.service
            if svc == 'builtin':
                raise exception.PolicyException(
                    "Builtins not handled by Z3 yet")
            lit_theory = theory if svc is None else self.theories[svc]
            translators = [
                self.type_registry.get_translator(str(arg_type.type))
                for arg_type in lit_theory.schema.types(name)]
            fullname = lit_theory.name + ":" + name
            try:
                z3func = self.relations[fullname]
                z3args = (compile_expr(arg, tr)
                          for (arg, tr) in six.moves.zip(literal.arguments,
                                                         translators))
                z3lit = z3func(*z3args)
                return (Z3OPT.Not(z3lit) if literal.negated
                        else z3lit)
            except KeyError:
                raise exception.PolicyException(
                    "Z3: Relation %s not registered" % fullname)

        z3head = compile_atom(head)
        z3body = [compile_atom(atom) for atom in body]
        # We give back variables explicitely and do not rely on declare_var and
        # abstract. Otherwise rules are cluttered with useless variables.
        return (z3vars, z3head, z3body)

    def compile_rule(self, theory, rule):
        # type: (Z3Theory, ast.Rule) -> None
        """compiles a single rule

        :param theory: the theory containing the rule
        :param rule: the rule to compile.
        """
        z3vars, z3head, z3body = self.compile_atoms(theory, rule.head,
                                                    rule.body)
        term1 = (z3head if z3body == []
                 else Z3OPT.Implies(Z3OPT.And(*z3body), z3head))
        term2 = term1 if z3vars == [] else Z3OPT.ForAll(z3vars, term1)
        self.context.rule(term2)

    def compile_query(self, theory, literal):
        # type: (Z3Theory, ast.Literal) -> z3.ExprRef
        """compiles a query litteral

        :param theory: theory used as the context of the query
        :param litteral: the query
        :returns: an existentially quantified litteral in Z3 format.
        """
        z3vars, z3head, _ = self.compile_atoms(theory, literal, [])
        return z3head if z3vars == [] else Z3OPT.Exists(z3vars, z3head)

    def compile_theory(self, theory):
        # type: (Z3Theory) -> None
        """Compiles all the rules of a theory

        :param theory: theory to compile. Will be marked clean after.
        """
        self.compile_facts(theory)
        for rules in six.itervalues(theory.rules.rules):
            for rule in rules:
                self.compile_rule(theory, rule)
        theory.dirty = False

    def compile_all(self):
        """Compile all Z3 theories"""
        self.relations = {}
        self.externals.clear()
        self.declare_tables()
        self.declare_external_tables()
        for theory in six.itervalues(self.z3theories):
            self.compile_theory(theory)
        self.last_compiled = time.time()

    def typecheck(self):
        """Typechecker for rules defined"""
        typer = typechecker.Typechecker(
            self.z3theories.values(), self.theories)
        typer.type_all()

    def inject(self, theoryname, tablename):
        # type: (str, str) -> None
        """Inject the values of an external table in the Z3 Context.

        Loops over the literal retrieved from a standard query.
        """
        theory = self.theories[theoryname]
        translators = [
            self.type_registry.get_translator(str(arg_type.type))
            for arg_type in theory.schema.types(tablename)]
        fullname = theory.name + ':' + tablename
        z3func = self.relations[fullname]
        for lit in retrieve(theory, tablename):
            z3args = (tr.to_z3(v.name, strict=True)
                      for (v, tr) in six.moves.zip(lit.arguments, translators))
            z3fact = z3func(*z3args)
            self.context.fact(z3fact)

    def synchronize_external(self):
        """Synchronize all external tables"""
        for (theoryname, tablename) in self.externals:
            self.inject(theoryname, tablename)



# Copyright (c) 2018 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import ipaddress
import json

from oslo_utils import uuidutils
import six


TypeNullabilityTuple = collections.namedtuple(
    'TypeNullabilityTuple', 'type nullable')


def nullable(marshal):
    '''decorator to make marshal function accept None value'''
    def func(cls, value):
        if value is None:
            return None
        else:
            return marshal(cls, value)
    return func


class UnqualifiedNameStr(abc.ABCMeta):
    '''metaclass to make str(Type) == Type'''
    def __str__(self):
        return self.__name__


@six.add_metaclass(UnqualifiedNameStr)
class CongressDataType(object):
    @classmethod
    @abc.abstractmethod
    def marshal(cls, value):
        '''Validate a value as valid for this type.

        :Raises ValueError: if the value is not valid for this type
        '''
        raise NotImplementedError

    @classmethod
    def least_ancestor(cls, target_types):
        '''Find this type's least ancestor among target_types

        This method helps a data consumer find the least common ancestor of
        this type among the types the data consumer supports.

        :param supported_types: iterable collection of types
        :returns: the subclass of CongressDataType which is the least ancestor
        '''
        target_types = frozenset(target_types)
        current_class = cls
        try:
            while current_class not in target_types:
                current_class = current_class._get_parent()
            return current_class
        except cls.CongressDataTypeNoParent:
            return None

    @classmethod
    def convert_to_ancestor(cls, value, ancestor_type):
        '''Convert this type's exchange value to ancestor_type's exchange value

        Generally there is no actual conversion because descendant type value
        is directly interpretable as ancestor type value. The only exception
        is the conversion from non-string descendents to string. This
        conversion is needed by Agnostic engine does not support boolean.

        .. warning:: undefined behavior if ancestor_type is not an ancestor of
                     this type.
        '''
        if ancestor_type == Str:
            return json.dumps(value)
        else:
            if cls.least_ancestor([ancestor_type]) is None:
                raise cls.CongressDataTypeHierarchyError
            else:
                return value

    @classmethod
    def _get_parent(cls):
        congress_parents = [parent for parent in cls.__bases__
                            if issubclass(parent, CongressDataType)]
        if len(congress_parents) == 1:
            return congress_parents[0]
        elif len(congress_parents) == 0:
            raise cls.CongressDataTypeNoParent(
                'No parent type found for {0}'.format(cls))
        else:
            raise cls.CongressDataTypeHierarchyError(
                'More than one parent type found for {0}: {1}'
                .format(cls, congress_parents))

    class CongressDataTypeNoParent(TypeError):
        pass

    class CongressDataTypeHierarchyError(TypeError):
        pass


class Scalar(CongressDataType):
    '''Most general type, emcompassing all JSON scalar values'''

    ACCEPTED_VALUE_TYPES = [
        six.string_types, six.text_type, six.integer_types, float, bool]

    @classmethod
    @nullable
    def marshal(cls, value):
        for type in cls.ACCEPTED_VALUE_TYPES:
            if isinstance(value, type):
                return value
        raise ValueError('Input value (%s) is of %s instead of one of the '
                         'expected types %s'
                         % (value, type(value), cls.ACCEPTED_VALUE_TYPES))


class Str(Scalar):

    @classmethod
    @nullable
    def marshal(cls, value):
        if not isinstance(value, six.string_types):
            raise ValueError('Input value (%s) is of %s instead of expected %s'
                             % (value, type(value), six.string_types))
        return value


class Bool(Scalar):

    @classmethod
    @nullable
    def marshal(cls, value):
        if not isinstance(value, bool):
            raise ValueError('Input value (%s) is of %s instead of expected %s'
                             % (value, type(value), bool))
        return value


class Int(Scalar):

    @classmethod
    @nullable
    def marshal(cls, value):
        if isinstance(value, int):
            return value
        elif isinstance(value, float) and value.is_integer():
            return int(value)
        else:
            raise ValueError('Input value (%s) is of %s instead of expected %s'
                             ' or %s' % (value, type(value), int, float))


class Float(Scalar):

    @classmethod
    @nullable
    def marshal(cls, value):
        if isinstance(value, float):
            return value
        elif isinstance(value, int):
            return float(value)
        else:
            raise ValueError('Input value (%s) is of %s instead of expected %s'
                             ' or %s' % (value, type(value), int, float))


class UUID(Str):

    @classmethod
    @nullable
    def marshal(cls, value):
        if uuidutils.is_uuid_like(value):
            return value
        else:
            raise ValueError('Input value (%s) is not an UUID' % value)


class IPAddress(Str):

    @classmethod
    @nullable
    def marshal(cls, value):
        try:
            return str(ipaddress.IPv4Address(six.text_type(value)))
        except ipaddress.AddressValueError:
            try:
                ipv6 = ipaddress.IPv6Address(six.text_type(value))
                if ipv6.ipv4_mapped:
                    return str(ipv6.ipv4_mapped)
                else:
                    return str(ipv6)
            except ipaddress.AddressValueError:
                raise ValueError('Input value (%s) is not interprable '
                                 'as an IP address' % value)


class IPNetwork(Str):

    @classmethod
    @nullable
    def marshal(cls, value):
        try:
            return str(ipaddress.ip_network(six.text_type(value)))
        except ValueError:
            raise ValueError('Input value (%s) is not interprable '
                             'as an IP network' % value)


@six.add_metaclass(abc.ABCMeta)
class CongressTypeFiniteDomain(object):
    '''Abstract base class for a Congress type of bounded domain.

    Each type inheriting from this class must have a class variable DOMAIN
    which is a frozenset of the set of values allowed in the type.
    '''
    pass


def create_congress_enum_type(class_name, enum_items, base_type,
                              catch_all_default_value=None):
    '''Return a sub-type of base_type

    representing a value of type base_type from a fixed, finite domain.
    :param enum_items: collection of items forming the domain
    :param catch_all_default_value: value to use for any value outside the
    domain. Defaults to None to disallow any avy value outside the domain.
    '''
    domain = set(enum_items)
    if catch_all_default_value is not None:
        domain.add(catch_all_default_value)

    for item in domain:
        if not base_type.marshal(item) == item:
            raise ValueError

    class NewType(base_type, CongressTypeFiniteDomain):
        DOMAIN = domain
        CATCH_ALL_DEFAULT_VALUE = catch_all_default_value

        @classmethod
        @nullable
        def marshal(cls, value):
            if value not in cls.DOMAIN:
                if cls.CATCH_ALL_DEFAULT_VALUE is None:
                    raise ValueError(
                        'Input value (%s) is not in the expected domain of '
                        'values %s' % (value, cls.DOMAIN))
                else:
                    return cls.CATCH_ALL_DEFAULT_VALUE
            return value

    NewType.__name__ = class_name
    return NewType


class TypesRegistry(object):
    _type_name_to_type_class = {}

    @classmethod
    def register(cls, type_class):
        # skip if type already registered
        if not issubclass(type_class, Scalar):
            raise TypeError('Attempted to register a type which is not a '
                            'subclass of the top type %s.' % Scalar)
        elif str(type_class) in cls._type_name_to_type_class:
            if type_class == cls._type_name_to_type_class[str(type_class)]:
                pass  # type already registered
            else:  # conflicting types with same name
                raise Exception('Attempted to register new type with the same '
                                'name \'%s\' as previously registered type.' %
                                type_class)
        else:  # register new type
            cls._type_name_to_type_class[str(type_class)] = type_class

    @classmethod
    def type_class(cls, type_name):
        return cls._type_name_to_type_class[type_name]


TYPES = [Scalar, Str, Bool, Int, Float, IPAddress, IPNetwork]

for type_class in TYPES:
    TypesRegistry.register((type_class))



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add datasource data persistence

Revision ID: 01e78af70b91
Revises: 3cee191c4f84
Create Date: 2016-07-29 17:02:40.026610

"""

# revision identifiers, used by Alembic.
revision = '01e78af70b91'
down_revision = '3cee191c4f84'

from alembic import op
import sqlalchemy as sa


def upgrade():
    if op.get_bind().engine.dialect.name == 'mysql':
        # NOTE: Specify a length long enough to choose MySQL
        # LONGTEXT
        text_type = sa.Text(length=1000000000)
    else:
        text_type = sa.Text()

    op.create_table(
        'dstabledata',
        sa.Column('ds_id', sa.String(length=36), nullable=False),
        sa.Column('tablename', sa.String(length=255), nullable=False),
        sa.Column('tabledata', text_type, nullable=False),
        sa.PrimaryKeyConstraint('ds_id', 'tablename'),
        mysql_engine='InnoDB')


def downgrade():
    op.drop_table('dstabledata')



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Unit Tests for Congress agent RPC
"""

from congress.cfg_validator.agent import rpc
from congress.tests import base_rpc


class TestValidatorDriverApi(base_rpc.BaseTestRpcClient):
    """Unit tests for RPC on the driver"""

    def test_process_templates_hashes(self):
        "unit test for process_templates_hashes"
        rpcapi = rpc.ValidatorDriverClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'process_templates_hashes',
            rpc_method='call',
            hashes='fake_hashes',
            host='host'
        )

    def test_process_configs_hashes(self):
        "unit test for process_configs_hashes"
        rpcapi = rpc.ValidatorDriverClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'process_configs_hashes',
            rpc_method='call',
            hashes='fake_hashes',
            host='host'
        )



# Copyright (c) 2018 VMware Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from datetime import datetime
from datetime import timedelta
import mock
import time

from congress.datasources import vitrage_driver
from congress.tests import base


class TestVitrageDriver(base.TestCase):

    def setUp(self):
        super(TestVitrageDriver, self).setUp()
        self.vitrage = vitrage_driver.VitrageDriver('test-vitrage')

    @mock.patch.object(vitrage_driver.VitrageDriver, 'publish')
    def test_webhook_alarm_activate(self, mocked_publish):
        test_payload = {
            "notification": "vitrage.alarm.activate",
            "payload": {
                "vitrage_id": "2def31e9-6d9f-4c16-b007-893caa806cd4",
                "resource": {
                    "vitrage_id": "437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac",
                    "name": "app-1-server-1-jz6qvznkmnif",
                    "update_timestamp": "2018-01-22 10:00:34.327142+00:00",
                    "vitrage_category": "RESOURCE",
                    "vitrage_operational_state": "OK",
                    "vitrage_type": "nova.instance",
                    "project_id": "8f007e5ba0944e84baa6f2a4f2b5d03a",
                    "id": "9b7d93b9-94ec-41e1-9cec-f28d4f8d702c"},
                "update_timestamp": "2018-01-22T10:00:34Z",
                "vitrage_category": "ALARM",
                "state": "Active",
                "vitrage_type": "vitrage",
                "vitrage_operational_severity": "WARNING",
                "name": "Instance memory performance degraded"}}

        self.vitrage._webhook_handler(test_payload)

        # check receive_timestamp
        self.assertTrue((
            abs(datetime.utcnow() -
                datetime.strptime(next(iter(self.vitrage.state['alarms']))[6],
                                  vitrage_driver.TIMESTAMP_FORMAT))
            <= timedelta(seconds=1)))

        self.assertEqual(1, len(self.vitrage.state['alarms']))

        expected_rows = set([(u'Instance memory performance degraded',
                              u'Active',
                              u'vitrage',
                              u'WARNING',
                              u'2def31e9-6d9f-4c16-b007-893caa806cd4',
                              u'2018-01-22T10:00:34Z',
                              next(iter(self.vitrage.state['alarms']))[6],
                              u'app-1-server-1-jz6qvznkmnif',
                              u'9b7d93b9-94ec-41e1-9cec-f28d4f8d702c',
                              u'437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac',
                              u'8f007e5ba0944e84baa6f2a4f2b5d03a',
                              u'OK',
                              u'nova.instance')])
        self.assertEqual(self.vitrage.state['alarms'], expected_rows)

    @mock.patch.object(vitrage_driver.VitrageDriver, 'publish')
    def test_webhook_alarm_deactivate(self, mocked_publish):
        test_payload = {
            "notification": "vitrage.alarm.deactivate",
            "payload": {
                "vitrage_id": "2def31e9-6d9f-4c16-b007-893caa806cd4",
                "resource": {
                    "vitrage_id": "437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac",
                    "name": "app-1-server-1-jz6qvznkmnif",
                    "update_timestamp": "2018-01-22 11:00:34.327142+00:00",
                    "vitrage_category": "RESOURCE",
                    "vitrage_operational_state": "OK",
                    "vitrage_type": "nova.instance",
                    "project_id": "8f007e5ba0944e84baa6f2a4f2b5d03a",
                    "id": "9b7d93b9-94ec-41e1-9cec-f28d4f8d702c"},
                "update_timestamp": "2018-01-22T11:00:34Z",
                "vitrage_category": "ALARM",
                "state": "Inactive",
                "vitrage_type": "vitrage",
                "vitrage_operational_severity": "OK",
                "name": "Instance memory performance degraded"}}

        self.vitrage.state['alarms'] = set([(
            u'Instance memory performance degraded',
            u'Active',
            u'vitrage',
            u'WARNING',
            u'2def31e9-6d9f-4c16-b007-893caa806cd4',
            u'2018-01-22T10:00:34Z',
            u'2018-01-22T10:00:34Z',
            u'app-1-server-1-jz6qvznkmnif',
            u'9b7d93b9-94ec-41e1-9cec-f28d4f8d702c',
            u'437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac',
            u'8f007e5ba0944e84baa6f2a4f2b5d03a',
            u'OK',
            u'nova.instance')])
        self.vitrage._webhook_handler(test_payload)

        self.assertEqual(1, len(self.vitrage.state['alarms']))

        expected_rows = set([(u'Instance memory performance degraded',
                              u'Inactive',
                              u'vitrage',
                              u'OK',
                              u'2def31e9-6d9f-4c16-b007-893caa806cd4',
                              u'2018-01-22T11:00:34Z',
                              next(iter(self.vitrage.state['alarms']))[6],
                              u'app-1-server-1-jz6qvznkmnif',
                              u'9b7d93b9-94ec-41e1-9cec-f28d4f8d702c',
                              u'437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac',
                              u'8f007e5ba0944e84baa6f2a4f2b5d03a',
                              u'OK',
                              u'nova.instance')])
        self.assertEqual(self.vitrage.state['alarms'], expected_rows)

    @mock.patch.object(vitrage_driver.VitrageDriver, 'publish')
    def test_webhook_alarm_cleanup(self, mocked_publish):
        self.vitrage = vitrage_driver.VitrageDriver(
            'test-vitrage',
            args={'hours_to_keep_alarm': 1 / 3600})  # set to 1 sec for test

        test_payload = {
            "notification": "vitrage.alarm.activate",
            "payload": {
                "vitrage_id": "2def31e9-6d9f-4c16-b007-893caa806cd4",
                "resource": {
                    "vitrage_id": "437f1f4c-ccce-40a4-ac62-1c2f1fd9f6ac",
                    "name": "app-1-server-1-jz6qvznkmnif",
                    "update_timestamp": "2018-01-22 10:00:34.327142+00:00",
                    "vitrage_category": "RESOURCE",
                    "vitrage_operational_state": "OK",
                    "vitrage_type": "nova.instance",
                    "project_id": "8f007e5ba0944e84baa6f2a4f2b5d03a",
                    "id": "9b7d93b9-94ec-41e1-9cec-f28d4f8d702c"},
                "update_timestamp": "2018-01-22T10:00:34Z",
                "vitrage_category": "ALARM",
                "state": "Active",
                "vitrage_type": "vitrage",
                "vitrage_operational_severity": "WARNING",
                "name": "Instance memory performance degraded"}}

        self.vitrage._webhook_handler(test_payload)

        self.assertEqual(1, len(self.vitrage.state['alarms']))
        time.sleep(3)
        self.assertEqual(0, len(self.vitrage.state['alarms']))



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

from congress.datasources import neutronv2_qos_driver
from congress.tests import base
from congress.tests import helper


class TestNeutronV2QosDriver(base.TestCase):

    def setUp(self):
        super(TestNeutronV2QosDriver, self).setUp()
        self.neutron_client_p = mock.patch(
            "neutronclient.v2_0.client.Client")
        self.neutron_client_p.start()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['client'] = mock.MagicMock()
        self.driver = neutronv2_qos_driver.NeutronV2QosDriver(args=args)

        self.mock_ports = {'ports': [
            {u'admin_state_up': True,
             u'allowed_address_pairs': [],
             u'binding:host_id': None,
             u'binding:vif_details': {u'port_filter': True},
             u'binding:vif_type': u'ovs',
             u'binding:vnic_type': u'normal',
             u'device_id': u'f42dc4f1-f371-48cc-95be-cf1b97112ab8',
             u'device_owner': u'network:router_gateway',
             u'fixed_ips': [
                 {u'ip_address': u'1.1.1.2',
                  u'subnet_id': u'10d20df9-e8ba-4756-ba30-d573ceb2e99a'}],
             u'id': u'04627c85-3553-436c-a7c5-0a64f5b87bb9',
             u'mac_address': u'fa:16:3e:f3:19:e5',
             u'name': u'',
             u'network_id': u'ecdea1af-7197-43c8-b3b0-34d90f72a2a8',
             u'port_security_enabled': False,
             u'security_groups': [],
             u'status': u'DOWN',
             u'tenant_id': u''},
            {u'admin_state_up': True,
             u'allowed_address_pairs': [],
             u'binding:host_id': None,
             u'binding:vif_details': {u'port_filter': True},
             u'binding:vif_type': u'ovs',
             u'binding:vnic_type': u'normal',
             u'device_id': u'f42dc4f1-f371-48cc-95be-cf1b97112ab8',
             u'device_owner': u'network:router_interface',
             u'fixed_ips': [
                 {u'ip_address': u'169.254.169.253',
                  u'subnet_id': u'aa9ad4f7-baf0-4a41-85c3-1cc8a3066db6'}],
             u'id': u'87f8933a-9582-48d8-ad16-9abf6e545002',
             u'mac_address': u'fa:16:3e:b7:78:e8',
             u'name': u'',
             u'network_id': u'6743ff85-2cfd-48a7-9d3f-472cd418783e',
             u'port_security_enabled': False,
             u'security_groups': [],
             u'status': u'DOWN',
             u'tenant_id': u''},
            {u'admin_state_up': True,
             u'allowed_address_pairs': [],
             u'binding:host_id': None,
             u'binding:vif_details': {u'port_filter': True},
             u'binding:vif_type': u'ovs',
             u'binding:vnic_type': u'normal',
             u'device_id': u'f42dc4f1-f371-48cc-95be-cf1b97112ab8',
             u'device_owner': u'network:router_interface',
             u'fixed_ips': [
                 {u'ip_address': u'10.0.0.1',
                  u'subnet_id': u'3c0eb3a3-4d16-4b1b-b327-44417182d0bb'}],
             u'id': u'c58c3246-6c2e-490a-b4d9-3b8d5191b465',
             u'mac_address': u'fa:16:3e:08:31:6e',
             u'name': u'',
             u'network_id': u'63ce8fbb-12e9-4ecd-9b56-1bbf8b51217d',
             u'port_security_enabled': False,
             u'security_groups': [],
             u'status': u'DOWN',
             u'tenant_id': u'feee0a965cc34274917fb753623dd57d'},
            {u'admin_state_up': True,
             u'allowed_address_pairs': [],
             u'binding:host_id': None,
             u'binding:vif_details': {u'port_filter': True},
             u'binding:vif_type': u'ovs',
             u'binding:vnic_type': u'normal',
             u'device_id': u'',
             u'device_owner': u'',
             u'fixed_ips': [
                 {u'ip_address': u'10.0.0.2',
                  u'subnet_id': u'3c0eb3a3-4d16-4b1b-b327-44417182d0bb'}],
             u'id': u'eb50003b-a081-4533-92aa-1cbd97f526a8',
             u'mac_address': u'fa:16:3e:af:56:fa',
             u'name': u'',
             u'network_id': u'63ce8fbb-12e9-4ecd-9b56-1bbf8b51217d',
             u'port_security_enabled': True,
             u'security_groups': [u'e0239062-4243-4798-865f-7055f03786d6'],
             u'qos_policies': [u'be50b732-4508-4a94-9c3c-8dc4b96a2b43'],
             u'status': u'DOWN',
             u'tenant_id': u'feee0a965cc34274917fb753623dd57d'}]}

        self.mock_qos_policies = {'policies': [
            {u'name': u'ysm',
             u'rules': [
                 {u'max_kbps': 100,
                  u'direction': u'egress',
                  u'qos_policy_id': u'be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                  u'type': u'bandwidth_limit',
                  u'id': u'9daaa87a-5441-49ef-8f25-2810d37c3a60',
                  u'max_burst_kbps': 500},
                 {u'dscp_mark': 10,
                  u'type': u'dscp_marking',
                  u'id': u'6be91937-b9ec-4209-a430-0c2694df1095',
                  u'qos_policy_id': u'be50b732-4508-4a94-9c3c-8dc4b96a2b43'},
                 {u'id': u'015f3dc8-7d3e-4598-8996-0597328c4db5',
                  u'direction': u'egress',
                  u'type': u'minimum_bandwidth',
                  u'qos_policy_id': u'be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                  u'min_kbps': 100}],
             u'tenant_id': u'feee0a965cc34274917fb753623dd57d',
             u'is_default': False,
             u'shared': False,
             u'project_id': u'feee0a965cc34274917fb753623dd57d',
             u'id': u'be50b732-4508-4a94-9c3c-8dc4b96a2b43',
             u'description': u''}]}

        self.expected_state = {
            'ports': set([(u'04627c85-3553-436c-a7c5-0a64f5b87bb9',),
                          (u'87f8933a-9582-48d8-ad16-9abf6e545002',),
                          (u'c58c3246-6c2e-490a-b4d9-3b8d5191b465',),
                          (u'eb50003b-a081-4533-92aa-1cbd97f526a8',)]),
            'qos_policy_port_bindings':
                set([('eb50003b-a081-4533-92aa-1cbd97f526a8',
                      'be50b732-4508-4a94-9c3c-8dc4b96a2b43')]),
            'policies':
                set([('be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                      'feee0a965cc34274917fb753623dd57d',
                      'ysm',
                      '',
                      False)]),
            'rules':
                set([('be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                      '015f3dc8-7d3e-4598-8996-0597328c4db5',
                      100,
                      'egress',
                      'minimum_bandwidth',
                      None,
                      None,
                      None),
                     ('be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                      '6be91937-b9ec-4209-a430-0c2694df1095',
                      None,
                      None,
                      'dscp_marking',
                      10,
                      None,
                      None),
                     ('be50b732-4508-4a94-9c3c-8dc4b96a2b43',
                      '9daaa87a-5441-49ef-8f25-2810d37c3a60',
                      None,
                      'egress',
                      'bandwidth_limit',
                      None,
                      500,
                      100)])}

    def test_update_from_datasource(self):
        with base.nested(
            mock.patch.object(self.driver.neutron,
                              "list_ports",
                              return_value=self.mock_ports),
            mock.patch.object(self.driver.neutron,
                              "list_qos_policies",
                              return_value=self.mock_qos_policies)):
            self.driver.update_from_datasource()
            self.assertEqual(self.expected_state, self.driver.state)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import glanceclient.v2.client as glclient
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


class GlanceV2Driver(datasource_driver.PollingDataSourceDriver,
                     datasource_driver.ExecutionDriver):

    IMAGES = "images"
    TAGS = "tags"

    value_trans = {'translation-type': 'VALUE'}
    images_translator = {
        'translation-type': 'HDICT',
        'table-name': IMAGES,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'UUID of image',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'The image status',
              'translator': value_trans},
             {'fieldname': 'name',
              'desc': 'Image Name', 'translator': value_trans},
             {'fieldname': 'container_format',
              'desc': 'The container format of image',
              'translator': value_trans},
             {'fieldname': 'created_at',
              'desc': 'The date and time when the resource was created',
              'translator': value_trans},
             {'fieldname': 'updated_at',
              'desc': 'The date and time when the resource was updated.',
              'translator': value_trans},
             {'fieldname': 'disk_format',
              'desc': 'The disk format of the image.',
              'translator': value_trans},
             {'fieldname': 'owner',
              'desc': 'The ID of the owner or tenant of the image',
              'translator': value_trans},
             {'fieldname': 'protected',
              'desc': 'Indicates whether the image can be deleted.',
              'translator': value_trans},
             {'fieldname': 'min_ram',
              'desc': 'minimum amount of RAM in MB required to boot the image',
              'translator': value_trans},
             {'fieldname': 'min_disk',
              'desc': 'minimum disk size in GB required to boot the image',
              'translator': value_trans},
             {'fieldname': 'checksum', 'desc': 'Hash of the image data used',
              'translator': value_trans},
             {'fieldname': 'size',
              'desc': 'The size of the image data, in bytes.',
              'translator': value_trans},
             {'fieldname': 'file',
              'desc': 'URL for the virtual machine image file',
              'translator': value_trans},
             {'fieldname': 'kernel_id', 'desc': 'kernel id',
              'translator': value_trans},
             {'fieldname': 'ramdisk_id', 'desc': 'ramdisk id',
              'translator': value_trans},
             {'fieldname': 'schema',
              'desc': 'URL for schema of the virtual machine image',
              'translator': value_trans},
             {'fieldname': 'visibility', 'desc': 'The image visibility',
              'translator': value_trans},
             {'fieldname': 'tags',
              'translator': {'translation-type': 'LIST',
                             'table-name': TAGS,
                             'val-col': 'tag',
                             'val-col-desc': 'List of image tags',
                             'parent-key': 'id',
                             'parent-col-name': 'image_id',
                             'parent-key-desc': 'UUID of image',
                             'translator': value_trans}})}

    TRANSLATORS = [images_translator]

    def __init__(self, name='', args=None):
        super(GlanceV2Driver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        session = ds_utils.get_keystone_session(self.creds)
        self.glance = glclient.Client(session=session)
        self.add_executable_client_methods(self.glance, 'glanceclient.v2.')
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'glancev2'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack Images aka Glance.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        images_method = lambda: self._translate_images(
            {'images': self.glance.images.list()})
        self.add_update_method(images_method, self.images_translator)

    @ds_utils.update_state_on_changed(IMAGES)
    def _translate_images(self, obj):
        """Translate the images represented by OBJ into tables."""
        LOG.debug("IMAGES: %s", str(dict(obj)))
        row_data = GlanceV2Driver.convert_objs(
            obj['images'], GlanceV2Driver.images_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.glance, action, action_args)



# Copyright (c) 2014 Styra, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from congress.api import base as api_base
from congress.policy_engines import agnostic
from congress.tests.api import base as tests_api_base
from congress.tests import base
from congress.tests import helper


class TestDseRuntime(base.SqlTestCase):
    @mock.patch('congress.db.db_policy_rules.get_policy_rules')
    def test_enable_schema(self, patched_persisted_rules):
        class TestRule(object):
            def __init__(self, id, name, rule_str,
                         policy_name, comment=None):
                self.id = id
                self.name = name
                self.rule = rule_str
                self.policy_name = policy_name
                self.comment = comment

        persisted_rule = [
            TestRule('rule-id', 'rule-name',
                     "p(x):-nova:services(id=x)", 'classification'),
            ]
        patched_persisted_rules.return_value = persisted_rule

        services = tests_api_base.setup_config()
        engine2 = services['engine']
        node = services['node']

        node.invoke_service_rpc = mock.MagicMock()
        node.invoke_service_rpc.return_value = [
            0, [['id1', 'name1', 'status1'], ['id2', 'name2', 'status2']]]

        # loaded rule is disabled
        subscriptions = engine2.subscription_list()
        self.assertEqual([], subscriptions)

        nova_schema = {
            'services': [
                {'name': 'id', 'desc': 'id of the service'},
                {'name': 'name', 'desc': 'name of the service'},
                {'name': 'status', 'desc': 'status of the service'}]}

        engine2.initialize_datasource('nova', nova_schema)
        # loaded rule is enabled and subscribes the table
        subscriptions = engine2.subscription_list()
        self.assertEqual([('nova', 'services')], subscriptions)


class TestAgnostic(base.TestCase):
    def test_receive_data_no_sequence_num(self):
        '''Test receiving data without sequence numbers'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        # initialize with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=None, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3], [4]], []], seqnum=None, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(2) p(3) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

        # remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[], [[2], [4]]], seqnum=None, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(3)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add & remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[4]], [[3]]], seqnum=None, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

        # re-initialize with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=None, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

    def test_receive_data_in_order(self):
        '''Test receiving data with sequence numbers, in order'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        # initialize with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=0, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3], [4]], []], seqnum=1, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(2) p(3) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

        # remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[], [[2], [4]]], seqnum=2, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(3)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add & remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[4]], [[3]]], seqnum=3, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

        # re-initialize with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=4, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

    def test_receive_data_out_of_order(self):
        '''Test receiving data with sequence numbers, out of order'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        # update with lower seqnum than init snapshot is ignored
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[10]], []], seqnum=3, is_snapshot=False)

        # add & remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[4]], [[3]]], seqnum=7, is_snapshot=False)
        actual = run.select('p(x)')
        correct = ''
        self.assertTrue(helper.db_equal(actual, correct))

        # remove data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[], [[2], [4]]], seqnum=6, is_snapshot=False)
        actual = run.select('p(x)')
        correct = ''
        self.assertTrue(helper.db_equal(actual, correct))

        # add data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3], [4]], []], seqnum=5, is_snapshot=False)
        actual = run.select('p(x)')
        correct = ''
        self.assertTrue(helper.db_equal(actual, correct))

        # initialize with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=4, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

    def test_receive_data_arbitrary_start(self):
        '''Test receiving data with arbitrary starting sequence number'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=1234, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

    def test_receive_data_duplicate_sequence_number(self):
        '''Test receiving data with duplicate sequence number

        Only one message (arbitrary) should be processed.
        '''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        # send three updates with the same seqnum
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[1]], []], seqnum=1, is_snapshot=False)
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[2]], []], seqnum=1, is_snapshot=False)
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3]], []], seqnum=1, is_snapshot=False)

        # start with empty data
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[], seqnum=0, is_snapshot=True)

        # exactly one of the three updates should be applied
        actual = run.select('p(x)')
        correct1 = 'p(1)'
        correct2 = 'p(2)'
        correct3 = 'p(3)'
        self.assertTrue(
            helper.db_equal(actual, correct1) or
            helper.db_equal(actual, correct2) or
            helper.db_equal(actual, correct3))

    def test_receive_data_sequence_number_max_int(self):
        '''Test receiving data when sequence number goes over max int'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1], [2]], seqnum=sys.maxsize, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[], [[2]]], seqnum=sys.maxsize + 1, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1)'
        self.assertTrue(helper.db_equal(actual, correct))

        # test out-of-sequence update ignored
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[2]], []], seqnum=sys.maxsize, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1)'
        self.assertTrue(helper.db_equal(actual, correct))

        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[4]], []], seqnum=sys.maxsize + 3, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1)'
        self.assertTrue(helper.db_equal(actual, correct))

        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3]], []], seqnum=sys.maxsize + 2, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(3) p(4)'
        self.assertTrue(helper.db_equal(actual, correct))

    def test_receive_data_multiple_tables(self):
        '''Test receiving data with sequence numbers, multiple tables'''
        run = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
        run.create_policy('datasource1')

        # initialize p with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[1]], seqnum=0, is_snapshot=True)
        actual = run.select('p(x)')
        correct = 'p(1)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add data to p
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[2]], []], seqnum=1, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(2)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add data to q
        run.receive_data_sequenced(
            publisher='datasource1', table='q',
            data=[[[2]], []], seqnum=1, is_snapshot=False)
        actual = run.select('q(x)')
        correct = ''  # does not apply until initialize
        self.assertTrue(helper.db_equal(actual, correct))

        # initialize q with full table
        run.receive_data_sequenced(
            publisher='datasource1', table='q',
            data=[[1]], seqnum=0, is_snapshot=True)
        actual = run.select('q(x)')
        correct = 'q(1) q(2)'  # both initial data and preceding update applied
        self.assertTrue(helper.db_equal(actual, correct))

        # add data to q
        run.receive_data_sequenced(
            publisher='datasource1', table='q',
            data=[[[3]], []], seqnum=2, is_snapshot=False)
        actual = run.select('q(x)')
        correct = 'q(1) q(2) q(3)'
        self.assertTrue(helper.db_equal(actual, correct))

        # add data to p
        run.receive_data_sequenced(
            publisher='datasource1', table='p',
            data=[[[3]], []], seqnum=2, is_snapshot=False)
        actual = run.select('p(x)')
        correct = 'p(1) p(2) p(3)'
        self.assertTrue(helper.db_equal(actual, correct))

    # TODO(ekcs): receive data multiple publishers



#    Copyright 2018 Orange
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Type translators between Congress and Z3."""
import abc
import six

from congress import data_types
from congress import exception

try:
    import z3
    Z3_AVAILABLE = True
except ImportError:
    Z3_AVAILABLE = False
    z3 = None

MYPY = False
if MYPY:
    # pylint: disable = unused-import
    from typing import Any, Union, List, Optional  # noqa


Z3OPT = z3


@six.add_metaclass(abc.ABCMeta)
class Z3Type(object):
    """Translate Openstack values to Z3"""

    def __init__(self, name, type_instance):
        self.name = name
        self.type_instance = type_instance

    @abc.abstractmethod
    def to_z3(self, val, strict=False):
        # type: (Any) -> z3.ExprRef
        """Transforms a value from OpenStack in a Z3 value"""
        raise NotImplementedError

    @abc.abstractmethod
    def to_os(self, val):
        # type: (z3.ExprRef) -> Any
        """Transforms a value from Z3 back to python"""
        raise NotImplementedError

    def type(self):
        # type: () -> z3.SortRef
        """Gives back the Z3 type"""
        return self.type_instance

    def reset(self):
        """Reset internal state of type transformer"""
        pass


class BoolType(Z3Type):
    """Transcode boolean in Z3"""

    def __init__(self):
        super(BoolType, self).__init__(u'Bool', z3.BoolSort())

    def to_z3(self, val, strict=False):
        return z3.BoolVal(val)

    def to_os(self, val):
        return val.decl().kind() == z3.Z3_OP_TRUE


class StringType(Z3Type):
    """Transcode strings in Z3"""

    def __init__(self, name, size=16):
        super(StringType, self).__init__(name, z3.BitVecSort(size))
        self.map = {}
        self.back = {}

    def to_z3(self, val, strict=False):
        if val in self.map:
            return self.map[val]
        code = len(self.map)
        bvect = z3.BitVecVal(code, self.type_instance)
        self.map[val] = bvect
        self.back[code] = val
        return bvect

    def to_os(self, val):
        return self.back[val.as_long()]

    def reset(self):
        self.map = {}
        self.back = {}


class FiniteType(StringType):
    """Z3 Coding for data_types with a finite number of elements

    This is the counterpart to data_types.CongressTypeFiniteDomain.
    """

    def __init__(self, name, domain):
        size = (len(domain) + 1).bit_length()
        super(FiniteType, self).__init__(name, size)
        self.domain = domain

    def to_z3(self, val, strict=False):
        if val in self.map:
            return self.map[val]
        if val not in self.domain and val is not None:
            if strict:
                raise exception.PolicyRuntimeException(
                    "Z3 Finite type: {} is not a value of {}".format(
                        val, self.name))
            else:
                val = '__OTHER__'
        code = len(self.map)
        bvect = z3.BitVecVal(code, self.type_instance)
        self.map[val] = bvect
        self.back[code] = val
        return bvect


class IntType(Z3Type):
    """Transcode numbers in Z3"""

    def __init__(self, name, size=32):
        super(IntType, self).__init__(name, z3.BitVecSort(size))
        self.map = {}
        self.back = {}

    def to_z3(self, val, strict=False):
        return z3.BitVecVal(val, self.type_instance)

    def to_os(self, val):
        return val.as_long()


class DummyType(Z3Type):
    """Dummy type when Z3 not available"""
    def to_z3(self, val, strict=False):
        pass

    def to_os(self, val):
        pass


class TypeRegistry(object):
    """A registry of Z3 types and their translators"""

    def __init__(self):
        self.type_dict = {}     # type: Dict[Str, Z3Type]
        self.top_type = DummyType('dummy', None)
        self.init()

    def register(self, typ):
        # type: (Z3Type) -> None
        """Registers a new Z3 type"""
        self.type_dict[typ.name] = typ

    def init(self):
        """Initialize the registry"""
        if Z3_AVAILABLE:
            self.top_type = StringType(u'Scalar', 34)
            for typ in [self.top_type, StringType(u'Str', 32),
                        IntType(u'Int', 32), BoolType(),
                        StringType('IPAddress', 32),
                        StringType('IPNetwork', 32),
                        StringType('UUID', 32)]:
                self.register(typ)

    def get_type(self, name):
        # type: (str) -> z3.SortRef
        """Return a Z3 type given a type name"""
        return self.get_translator(name).type()

    def get_translator(self, name):
        # type: (str) -> Z3Type
        """Return the translator for a given type name"""
        trans = self.type_dict.get(name, None)
        if trans is None:
            try:
                congress_type = data_types.TypesRegistry.type_class(name)
            except KeyError:
                raise exception.PolicyRuntimeException(
                    "Z3 typechecker: Unknown congress type {}".format(name))
            if issubclass(congress_type, data_types.CongressTypeFiniteDomain):
                trans = FiniteType(name, congress_type.DOMAIN)
                self.register(trans)
            else:
                raise exception.PolicyRuntimeException(
                    "Z3 typechecker: cannot handle type {}".format(name))
        return trans

    def reset(self):
        # type: () -> None
        """Reset the internal tables of all types"""
        for typ in six.itervalues(self.type_dict):
            typ.reset()


def z3_to_array(expr):
    # type: (z3.BoolRef) -> Union[bool, List[List[Any]]]
    """Compiles back a Z3 result to a matrix of values"""
    def extract(item):
        """Extract a row"""
        kind = item.decl().kind()
        if kind == z3.Z3_OP_AND:
            return [x.children()[1] for x in item.children()]
        elif kind == z3.Z3_OP_EQ:
            return [item.children()[1]]
        else:
            raise exception.PolicyRuntimeException(
                "Bad Z3 result not translatable {}: {}".format(expr, kind))
    kind = expr.decl().kind()

    if kind == z3.Z3_OP_OR:
        return [extract(item) for item in expr.children()]
    elif kind == z3.Z3_OP_AND:
        return [[item.children()[1] for item in expr.children()]]
    elif kind == z3.Z3_OP_EQ:
        return [[expr.children()[1]]]
    elif kind == z3.Z3_OP_FALSE:
        return False
    elif kind == z3.Z3_OP_TRUE:
        return True
    else:
        raise exception.PolicyRuntimeException(
            "Bad Z3 result not translatable {}: {}".format(expr, kind))



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

from oslo_config import cfg
from oslo_log import log as logging

from congress.api import action_model
from congress.api import application
from congress.api import base as api_base
from congress.api import datasource_model
from congress.api import library_policy_model
from congress.api import policy_model
from congress.api import router
from congress.api import row_model
from congress.api import rule_model
from congress.api import schema_model
from congress.api import status_model
from congress.api.system import driver_model
from congress.api import table_model
from congress.api import webhook_model
from congress.db import datasources as db_datasources
from congress.dse2 import datasource_manager as ds_manager
from congress.dse2 import dse_node
from congress import exception
from congress.library_service import library_service
from congress.policy_engines import agnostic

LOG = logging.getLogger(__name__)


def create2(node_id=None, bus_id=None, existing_node=None,
            policy_engine=True, datasources=True, api=True):
    """Get Congress up.

    Creates a DseNode if one is not provided and adds policy_engine,
    datasources, api to that node.

    :param: node_id is node_id of DseNode to be created
    :param: bus_id is partition_id of DseNode to be created
    :param: existing_node is a DseNode (optional; in lieu of previous 2 params)
    :param: policy_engine controls whether policy_engine is included
    :param: datasources controls whether datasources are included
    :param: api controls whether API is included
    :returns: DseNode
    """
    # create DseNode if existing_node not given
    if existing_node is None:
        assert (not (node_id is None or bus_id is None)),\
            'params node_id and bus_id required.'
        node = dse_node.DseNode(cfg.CONF, node_id, [], partition_id=bus_id)
    else:
        assert (node_id is None and bus_id is None),\
            'params node_id and bus_id must be None when existing_node given.'
        node = existing_node

    # create services as required
    services = {}

    # Load all configured drivers
    dse_node.DseNode.load_drivers()

    if datasources:
        LOG.info("Registering congress datasource services on node %s",
                 node.node_id)
        services['datasources'] = create_datasources(node)
        services['ds_manager'] = ds_manager.DSManagerService(
            api_base.DS_MANAGER_SERVICE_ID)
        node.register_service(services['ds_manager'])

    if policy_engine:
        LOG.info("Registering congress PolicyEngine service on node %s",
                 node.node_id)
        engine = create_policy_engine()
        services[api_base.ENGINE_SERVICE_ID] = engine
        node.register_service(engine)
        initialize_policy_engine(engine)

        # NOTE(ekcs): library service does not depend on policy engine,
        # it is placed on the same nodes as policy engine for convenience only
        LOG.info("Registering congress policy library service on node %s",
                 node.node_id)
        library = create_policy_library_service()
        services[api_base.LIBRARY_SERVICE_ID] = library
        node.register_service(library)

    if api:
        LOG.info("Registering congress API service on node %s", node.node_id)
        services['api'], services['api_service'] = create_api()
        node.register_service(services['api_service'])

    return services


def create_api():
    """Return service that encapsulates api logic for DSE2."""
    # ResourceManager inherits from DataService
    api_resource_mgr = application.ResourceManager()
    models = create_api_models(api_resource_mgr)
    router.APIRouterV1(api_resource_mgr, models)
    return models, api_resource_mgr


def create_api_models(bus):
    """Create all the API models and return as a dictionary for DSE2."""
    res = {}
    res['api-library-policy'] = library_policy_model.LibraryPolicyModel(
        'api-library-policy', bus=bus)
    res['api-policy'] = policy_model.PolicyModel('api-policy', bus=bus)
    res['api-rule'] = rule_model.RuleModel('api-rule', bus=bus)
    res['api-row'] = row_model.RowModel('api-row', bus=bus)
    res['api-datasource'] = datasource_model.DatasourceModel(
        'api-datasource', bus=bus)
    res['api-schema'] = schema_model.SchemaModel('api-schema', bus=bus)
    res['api-table'] = table_model.TableModel('api-table', bus=bus)
    res['api-webhook'] = webhook_model.WebhookModel('api-webhook', bus=bus)
    res['api-status'] = status_model.StatusModel('api-status', bus=bus)
    res['api-action'] = action_model.ActionsModel('api-action', bus=bus)
    res['api-system'] = driver_model.DatasourceDriverModel(
        'api-system', bus=bus)
    return res


def create_policy_engine():
    """Create policy engine and initialize it using the api models."""
    engine = agnostic.DseRuntime(api_base.ENGINE_SERVICE_ID)
    engine.debug_mode()  # should take this out for production
    return engine


def initialize_policy_engine(engine):
    """Initialize the policy engine using the API."""
    # Load policies from database
    engine.persistent_load_policies()
    engine.create_default_policies()
    engine.persistent_load_rules()


def create_policy_library_service():
    """Create policy library service."""
    library = library_service.LibraryService(api_base.LIBRARY_SERVICE_ID)
    # load library policies from file if none present in DB
    if len(library.get_policies(include_rules=False)) == 0:
        library.load_policies_from_files()
    return library


def create_datasources(bus):
    """Create and register datasource services ."""
    if cfg.CONF.delete_missing_driver_datasources:
        # congress server started with --delete-missing-driver-datasources
        bus.delete_missing_driver_datasources()

    datasources = db_datasources.get_datasources()
    services = []
    for ds in datasources:
        LOG.info("create configured datasource service %s.", ds.name)
        try:
            service = bus.create_datasource_service(ds)
            if service:
                bus.register_service(service)
                services.append(service)
        except exception.DriverNotFound:
            LOG.exception("Some datasources could not be loaded, start "
                          "congress server with "
                          "--delete-missing-driver-datasources option to "
                          "clean up stale datasources in DB.")
            sys.exit(1)
        except Exception:
            LOG.exception("datasource %s creation failed. %s service may not "
                          "be running.", ds.name, ds.driver)
    return services



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Schema change history

date: 2018-03-15
changes:

 - (incompatible) Removed the `hosts` table for OS hosts information because
   access to the information has been removed from the latest Nova API and
   client.
 - Added the `hypervisors` table for hypervisor information.

date: 2017-10-01
changes:

 - Added the `tags` table for server tags information.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import novaclient.client
from oslo_log import log as logging
import six

from congress import data_types
from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


LOG = logging.getLogger(__name__)


class NovaDriver(datasource_driver.PollingDataSourceDriver,
                 datasource_driver.ExecutionDriver):
    SERVERS = "servers"
    FLAVORS = "flavors"
    HYPERVISORS = "hypervisors"
    SERVICES = 'services'
    AVAILABILITY_ZONES = "availability_zones"
    TAGS = "tags"

    value_trans_str = ds_utils.typed_value_trans(data_types.Str)
    value_trans_bool = ds_utils.typed_value_trans(data_types.Bool)
    value_trans_int = ds_utils.typed_value_trans(data_types.Int)

    def safe_id(x):
        if isinstance(x, six.string_types):
            return x
        try:
            return x['id']
        except Exception:
            return str(x)

    servers_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVERS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the server',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'Name of the server',
              'translator': value_trans_str},
             {'fieldname': 'hostId', 'col': 'host_id',
              'desc': 'The UUID for the host', 'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'The server status',
              'translator': value_trans_str},
             {'fieldname': 'tenant_id', 'desc': 'The tenant ID',
              'translator': value_trans_str},
             {'fieldname': 'user_id',
              'desc': 'The user ID of the user who owns the server',
              'translator': value_trans_str},
             {'fieldname': 'image', 'col': 'image_id',
              'desc': 'Name or ID of image',
              'translator': {'translation-type': 'VALUE',
                             'extract-fn': safe_id}},
             {'fieldname': 'flavor', 'col': 'flavor_id',
              'desc': 'ID of the flavor',
              'translator': {'translation-type': 'VALUE',
                             'extract-fn': safe_id}},
             {'fieldname': 'OS-EXT-AZ:availability_zone', 'col': 'zone',
              'desc': 'The availability zone of host',
              'translator': value_trans_str},
             {'fieldname': 'OS-EXT-SRV-ATTR:hypervisor_hostname',
              'desc': ('The hostname of hypervisor where the server is '
                       'running'),
              'col': 'host_name', 'translator': value_trans_str},
             {'fieldname': 'tags',
              'translator': {'translation-type': 'LIST',
                             'table-name': TAGS,
                             'parent-key': 'id',
                             'parent-col-name': 'server_id',
                             'parent-key-desc': 'UUID of server',
                             'val-col': 'tag',
                             'val-col-desc': 'server tag string',
                             'translator': value_trans_str}})}

    flavors_translator = {
        'translation-type': 'HDICT',
        'table-name': FLAVORS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'ID of the flavor',
              'translator': value_trans_str},
             {'fieldname': 'name', 'desc': 'Name of the flavor',
              'translator': value_trans_str},
             {'fieldname': 'vcpus', 'desc': 'Number of vcpus',
              'translator': value_trans_int},
             {'fieldname': 'ram', 'desc': 'Memory size in MB',
              'translator': value_trans_int},
             {'fieldname': 'disk', 'desc': 'Disk size in GB',
              'translator': value_trans_int},
             {'fieldname': 'ephemeral', 'desc': 'Ephemeral space size in GB',
              'translator': value_trans_int},
             {'fieldname': 'rxtx_factor', 'desc': 'RX/TX factor',
              'translator': ds_utils.typed_value_trans(data_types.Float)})}

    hypervisors_translator = {
        'translation-type': 'HDICT',
        'table-name': HYPERVISORS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'hypervisor_hostname', 'desc': 'Hypervisor host',
              'translator': value_trans_str},
             {'fieldname': 'id', 'desc': 'hypervisori id',
              # untyped: depends on api microversion
              'translator': {'translation-type': 'VALUE'}},
             {'fieldname': 'state', 'desc': 'State of the hypervisor',
              'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'Status of the hypervisor',
              'translator': value_trans_str})}

    services_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVICES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'col': 'service_id', 'desc': 'Service ID',
              'translator': value_trans_int},
             {'fieldname': 'binary', 'desc': 'Service binary',
              'translator': value_trans_str},
             {'fieldname': 'host', 'desc': 'Host Name',
              'translator': value_trans_str},
             {'fieldname': 'zone', 'desc': 'Availability Zone',
              'translator': value_trans_str},
             {'fieldname': 'status', 'desc': 'Status of service',
              'translator': value_trans_str},
             {'fieldname': 'state', 'desc': 'State of service',
              'translator': value_trans_str},
             {'fieldname': 'updated_at', 'desc': 'Last updated time',
              'translator': value_trans_str},
             {'fieldname': 'disabled_reason', 'desc': 'Disabled reason',
              'translator': value_trans_str})}

    availability_zones_translator = {
        'translation-type': 'HDICT',
        'table-name': AVAILABILITY_ZONES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'zoneName', 'col': 'zone',
              'desc': 'Availability zone name', 'translator': value_trans_str},
             {'fieldname': 'zoneState', 'col': 'state',
              'desc': 'Availability zone state',
              'translator': value_trans_str})}

    TRANSLATORS = [servers_translator, flavors_translator, services_translator,
                   hypervisors_translator, availability_zones_translator]

    def __init__(self, name='', args=None):
        super(NovaDriver, self).__init__(name, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        session = ds_utils.get_keystone_session(self.creds)
        self.nova_client = novaclient.client.Client(
            version=self.creds.get('api_version', '2.26'), session=session)
        self.add_executable_method('servers_set_meta',
                                   [{'name': 'server',
                                    'description': 'server id'},
                                    {'name': 'meta-key1',
                                     'description': 'meta key 1'},
                                    {'name': 'meta-value1',
                                     'description': 'value for meta key1'},
                                    {'name': 'meta-keyN',
                                     'description': 'meta key N'},
                                    {'name': 'meta-valueN',
                                     'description': 'value for meta keyN'}],
                                   "A wrapper for servers.set_meta()")
        self.add_executable_client_methods(self.nova_client, 'novaclient.v2.')
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'nova'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack Compute aka nova.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['api_version'] = constants.OPTIONAL
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        servers_method = lambda: self._translate_servers(
            self.nova_client.servers.list(
                detailed=True, search_opts={"all_tenants": 1}))
        self.add_update_method(servers_method, self.servers_translator)

        flavors_method = lambda: self._translate_flavors(
            self.nova_client.flavors.list())
        self.add_update_method(flavors_method, self.flavors_translator)

        hypervisors_method = lambda: self._translate_hypervisors(
            self.nova_client.hypervisors.list())
        self.add_update_method(hypervisors_method,
                               self.hypervisors_translator)

        services_method = lambda: self._translate_services(
            self.nova_client.services.list())
        self.add_update_method(services_method, self.services_translator)

        az_method = lambda: self._translate_availability_zones(
            self.nova_client.availability_zones.list())
        self.add_update_method(az_method, self.availability_zones_translator)

    @ds_utils.update_state_on_changed(SERVERS)
    def _translate_servers(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.servers_translator)
        return row_data

    @ds_utils.update_state_on_changed(FLAVORS)
    def _translate_flavors(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.flavors_translator)
        return row_data

    @ds_utils.update_state_on_changed(HYPERVISORS)
    def _translate_hypervisors(self, obj):
        row_data = NovaDriver.convert_objs(
            obj,
            NovaDriver.hypervisors_translator)
        return row_data

    @ds_utils.update_state_on_changed(SERVICES)
    def _translate_services(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.services_translator)
        return row_data

    @ds_utils.update_state_on_changed(AVAILABILITY_ZONES)
    def _translate_availability_zones(self, obj):
        row_data = NovaDriver.convert_objs(
            obj,
            NovaDriver.availability_zones_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.nova_client, action, action_args)

    # "action" methods - to be used with "execute"
    def servers_set_meta(self, args):
        """A wrapper for servers.set_meta().

        'execute[p(x)]' doesn't take optional args at the moment.
        Therefore, this function translates the positional ARGS
        to optional args and call the servers.set_meta() api.
        :param: <list> args: expected server ID and pairs of meta
        data in positional args such as::

            {'positional': ['server_id', 'meta1', 'value1', 'meta2', 'value2']}

        Usage::

            execute[nova.servers_set_meta(svr_id, meta1, val1, meta2, val2) :-
                triggering_table(id)
        """
        action = 'servers.set_meta'
        positional_args = args.get('positional', [])
        if not positional_args:
            LOG.error('Args not found for servers_set_meta()')
            return

        # Strip off the server_id before converting meta data pairs
        server_id = positional_args.pop(0)
        meta_data = self._convert_args(positional_args)

        action_args = {'named': {'server': server_id,
                                 'metadata': meta_data}}
        self._execute_api(self.nova_client, action, action_args)



# Copyright (c) 2016 NTT All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import six

from congress.datasources import constants
from congress.datasources import datasource_driver


class DoctorDriver(datasource_driver.PushedDataSourceDriver):
    """A DataSource Driver for OPNFV Doctor project.

    This driver has a table for Doctor project's Inspector. Please check
    https://wiki.opnfv.org/display/doctor/Doctor+Home for the details
    about OPNFV Doctor project.

    To update the table, call Update row API.

    PUT /v1/data-sources/<the driver id>/tables/<table id>/rows

    For updating 'events' table, the request body should be following
    style. The request will replace all rows in the table with the body,
    which means if you update the table with [] it will clear the table.
    One {} object in the list represents one row of the table.

    request body::

        [
          {
            "time": "2016-02-22T11:48:55Z",
            "type": "compute.host.down",
            "details": {
                "hostname": "compute1",
                "status": "down",
                "monitor": "zabbix1",
                "monitor_event_id": "111"
            }
          },
          .....
        ]
    """

    value_trans = {'translation-type': 'VALUE'}

    def safe_id(x):
        if isinstance(x, six.string_types):
            return x
        try:
            return x['id']
        except Exception:
            return str(x)

    def flatten_events(row_events):
        flatten = []
        for event in row_events:
            details = event.pop('details')
            for k, v in details.items():
                event[k] = v
            flatten.append(event)
        return flatten

    events_translator = {
        'translation-type': 'HDICT',
        'table-name': 'events',
        'selector-type': 'DICT_SELECTOR',
        'objects-extract-fn': flatten_events,
        'field-translators':
            ({'fieldname': 'time', 'translator': value_trans},
             {'fieldname': 'type', 'translator': value_trans},
             {'fieldname': 'hostname', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'monitor', 'translator': value_trans},
             {'fieldname': 'monitor_event_id', 'translator': value_trans},)
        }

    TRANSLATORS = [events_translator]

    def __init__(self, name='', args=None):
        super(DoctorDriver, self).__init__(name, args=args)

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'doctor'
        result['description'] = ('Datasource driver that allows external '
                                 'systems to push data in accordance with '
                                 'OPNFV Doctor Inspector southbound interface '
                                 'specification.')
        result['config'] = {'persist_data': constants.OPTIONAL}
        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

sys.path.insert(0, os.path.abspath('../..'))
sys.path.insert(0, os.path.abspath('../'))
sys.path.insert(0, os.path.abspath('./'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.todo',
    'sphinxcontrib.apidoc',
    'oslo_config.sphinxext',
    'oslo_config.sphinxconfiggen',
]

# Don't use default openstack theme, for readthedocs
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:
    extensions.append('openstackdocstheme')

# openstackdocstheme options
repository_name = 'openstack/congress'
bug_project = 'congress'
bug_tag = ''
# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../congress'
apidoc_output_dir = 'api'
apidoc_excluded_paths = [
    'datalog/Python2/*',
    'datalog/Python3/*',
    'db/migration/alembic_migrations/*',
    'server/*',
    'tests/*',
    '/dse2/disabled_test_control_bus.py']

apidoc_separate_modules = True

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'congress'
copyright = u'2013, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of glob-style patterns that should be excluded when looking for
# source files. They are matched against the source file names relative to the
# source directory, using slashes as directory separators on all platforms.
exclude_patterns = ['api/congress.db.migration.alembic_migrations.*',
                    'api/congress.server.*']


# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['congress.']

autodoc_mock_imports = ['congress.datalog.Python2', 'congress.datalog.Python3',
                        'cloudfoundryclient', 'congress.dse',
                        'monascaclient']

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
html_theme = 'openstackdocs'
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

html_last_updated_fmt = '%Y-%m-%d %H:%M'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     '%s.tex' % project,
     u'%s Documentation' % project,
     u'OpenStack Foundation', 'manual'),
]

# Example configuration for intersphinx: refer to the Python standard library.
#intersphinx_mapping = {'http://docs.python.org/': None}

# -- Options for oslo_config.sphinxconfiggen ---------------------------------

config_generator_config_file = [
    ('../../etc/congress-config-generator.conf',
     '_static/congress'),
    ('../../etc/congress-agent-config-generator.conf',
     '_static/congress-agent')
]


[extensions]
todo_include_todos=True



# Copyright (c) 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

""" Base class for all API models."""

from __future__ import absolute_import

from oslo_config import cfg

ENGINE_SERVICE_ID = '__engine'
LIBRARY_SERVICE_ID = '__library'
DS_MANAGER_SERVICE_ID = '_ds_manager'


class APIModel(object):
    """Base Class for handling API requests."""

    def __init__(self, name, bus=None):
        self.name = name
        self.dse_long_timeout = cfg.CONF.dse.long_timeout
        self.action_retry_timeout = cfg.CONF.dse.execute_action_retry_timeout
        self.bus = bus

    # Note(thread-safety): blocking function
    def invoke_rpc(self, caller, name, kwargs, timeout=None):
            local = (caller is ENGINE_SERVICE_ID and
                     self.bus.node.service_object(
                         ENGINE_SERVICE_ID) is not None)
            return self.bus.rpc(
                caller, name, kwargs, timeout=timeout, local=local)



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from keystoneclient.v3 import client

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


class KeystoneV3Driver(datasource_driver.PollingDataSourceDriver,
                       datasource_driver.ExecutionDriver):
    # Table names
    USERS = "users"
    ROLES = "roles"
    PROJECTS = "projects"
    DOMAINS = "domains"

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    users_translator = {
        'translation-type': 'HDICT',
        'table-name': USERS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The ID for the user.',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'username, unique within domain',
              'translator': value_trans},
             {'fieldname': 'enabled', 'desc': 'user is enabled or not',
              'translator': value_trans},
             {'fieldname': 'default_project_id',
              'desc': 'ID of the default project for the user',
              'translator': value_trans},
             {'fieldname': 'domain_id',
              'desc': 'The ID of the domain for the user.',
              'translator': value_trans})}

    roles_translator = {
        'translation-type': 'HDICT',
        'table-name': ROLES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'role ID', 'translator': value_trans},
             {'fieldname': 'name', 'desc': 'role name',
              'translator': value_trans})}

    projects_translator = {
        'translation-type': 'HDICT',
        'table-name': PROJECTS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'enabled', 'desc': 'project is enabled or not',
              'translator': value_trans},
             {'fieldname': 'description', 'desc': 'project description',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'project name',
              'translator': value_trans},
             {'fieldname': 'domain_id',
              'desc': 'The ID of the domain for the project',
              'translator': value_trans},
             {'fieldname': 'id', 'desc': 'ID for the project',
              'translator': value_trans})}

    domains_translator = {
        'translation-type': 'HDICT',
        'table-name': DOMAINS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'enabled', 'desc': 'domain is enabled or disabled',
              'translator': value_trans},
             {'fieldname': 'description', 'desc': 'domain description',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'domain name',
              'translator': value_trans},
             {'fieldname': 'id', 'desc': 'domain ID',
              'translator': value_trans})}

    TRANSLATORS = [users_translator, roles_translator, projects_translator,
                   domains_translator]

    def __init__(self, name='', args=None):
        super(KeystoneV3Driver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        session = ds_utils.get_keystone_session(args)
        self.client = client.Client(session=session)
        self.add_executable_client_methods(self.client,
                                           'keystoneclient.v3.client')
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'keystonev3'
        result['description'] = ('Datasource driver that interfaces with '
                                 'keystone.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        users_method = lambda: self._translate_users(self.client.users.list())
        self.add_update_method(users_method, self.users_translator)

        roles_method = lambda: self._translate_roles(self.client.roles.list())
        self.add_update_method(roles_method, self.roles_translator)

        projects_method = lambda: self._translate_projects(
            self.client.projects.list())
        self.add_update_method(projects_method, self.projects_translator)

        domains_method = lambda: self._translate_domains(
            self.client.domains.list())
        self.add_update_method(domains_method, self.domains_translator)

    @ds_utils.update_state_on_changed(USERS)
    def _translate_users(self, obj):
        row_data = KeystoneV3Driver.convert_objs(
            obj, KeystoneV3Driver.users_translator)
        return row_data

    @ds_utils.update_state_on_changed(ROLES)
    def _translate_roles(self, obj):
        row_data = KeystoneV3Driver.convert_objs(
            obj, KeystoneV3Driver.roles_translator)
        return row_data

    @ds_utils.update_state_on_changed(PROJECTS)
    def _translate_projects(self, obj):
        row_data = KeystoneV3Driver.convert_objs(
            obj, KeystoneV3Driver.projects_translator)
        return row_data

    @ds_utils.update_state_on_changed(DOMAINS)
    def _translate_domains(self, obj):
        row_data = KeystoneV3Driver.convert_objs(
            obj, KeystoneV3Driver.domains_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.client, action, action_args)



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Test the configuration validator driver"""

import mock
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
import six
from testtools.content import text_content

from congress.datasources import cfgvalidator_driver
from congress.tests import base
from congress.tests import base_rpc
from congress.tests import helper

LOG = logging.getLogger(__name__)


# pylint: disable=protected-access
def _fake_conf():
    conf = mock.MagicMock()
    conf._namespace = 'ns'
    opt1 = mock.MagicMock()
    opt1.id_ = 'ho1'
    opt1.name = 'o1'
    opt1.type = types.String
    opt1.ns_id = 'ns'
    opt2 = mock.MagicMock()
    opt2.id_ = 'ho2'
    opt2.name = 'o2'
    opt2.type = types.String
    opt2.ns_id = 'ns'
    group = mock.MagicMock()
    group._opts = {'o2': {'opt': opt2}}
    conf._groups = {'g': group}
    conf._opts = {'o1': {'opt': opt1}}
    return conf


class TestCfgValidatorDriver(base.TestCase):
    """Test the configuration validator driver"""
    def setUp(self):
        super(TestCfgValidatorDriver, self).setUp()
        args = helper.datasource_openstack_args()
        with mock.patch('congress.datasources.cfgvalidator_driver.'
                        'ValidatorAgentClient',
                        spec=cfgvalidator_driver.ValidatorAgentClient) as agm:
            self.driver = cfgvalidator_driver.ValidatorDriver(args=args)
            self.agent_mock = agm
        self.driver.node = mock.MagicMock()

        for table in cfgvalidator_driver.ValidatorDriver.get_schema():
            self.driver.state[table] = set()

    def test_get_info(self):
        """Test info retrieval on datasource. Minimal requirements"""
        info = self.driver.get_datasource_info()
        self.assertIsNotNone(info['id'])
        self.assertIsNotNone(info['description'])
        self.assertIsNotNone(info['config'])

    def test_translate_type(self):
        """Test the translation of type"""
        cases = [
            {
                'inputs': ['lorem',
                           types.String(choices=['foo'], max_length=4)],
                'expected': {
                    cfgvalidator_driver.STR_TYPE:
                        (u'lorem', u'', 4, u'False', u'False', u'[\'foo\']')}
            },
            {
                'inputs': ['lorem', types.Integer(choices=[1], min=1, max=2)],
                'expected': {
                    cfgvalidator_driver.INT_TYPE: (u'lorem', 1, 2, u'[1]')}
            },
            {
                'inputs': ['lorem', types.Float(min=1, max=2)],
                'expected': {cfgvalidator_driver.FLOAT_TYPE: (u'lorem', 1, 2)}
            },
            {
                'inputs': ['lorem', types.List(item_type=types.Float(min=1))],
                'expected': {
                    cfgvalidator_driver.LIST_TYPE: (
                        u'lorem', u'Float', u'False'),
                    cfgvalidator_driver.FLOAT_TYPE: (u'lorem', 1, u''), }
            },
            {
                'inputs': ['lorem', types.URI(max_length=2, schemes=['HTTP'])],
                'expected': {
                    cfgvalidator_driver.URI_TYPE: (u'lorem', 2, u'[\'HTTP\']')}
            },
            {
                'inputs': ['lorem', types.Range(min=1, max=2)],
                'expected': {cfgvalidator_driver.RANGE_TYPE: (u'lorem', 1, 2)}
            },
        ]

        for case in cases:
            self.driver.translate_type(*case['inputs'])

        for case in cases:
            for table_name, expected in six.iteritems(case['expected']):
                table = self.driver.state[table_name]

                if expected:
                    self.assertIn(expected, table)

    def test_translate_host(self):
        """Test the translation of host"""
        cases = [
            ('lorem', 'ipsum', (u'lorem', u'ipsum')),
            (None, 'ipsum', None),
            ('', 'ipsum', None),
            ('lorem', None, (u'lorem', u'')),
            ('lorem', '', (u'lorem', u'')),
        ]

        for host_id, host_name, _ in cases:
            self.driver.translate_host(host_id, host_name)

        table = self.driver.state[cfgvalidator_driver.HOST]

        for _, _, expected in cases:
            if expected:
                self.assertIn(expected, table)

        expected_size = len(set([c[-1] for c in cases if c[-1]]))
        self.assertEqual(len(table), expected_size)

    def test_translate_file(self):
        """Test the translation of file"""
        cases = [
            ('lorem', 'ipsum', 'dolor', 'sit',
             (u'lorem', u'ipsum', u'dolor', u'sit')),
            ('lorem', 'ipsum', None, '', (u'lorem', u'ipsum', u'', u'')),
            ('lorem', 'ipsum', '', None, (u'lorem', u'ipsum', u'', u'')),
            (None, 'ipsum', 'dolor', 'sit', None),
            ('', 'ipsum', 'dolor', 'sit', None),
            ('lorem', '', 'dolor', 'sit', None),
            ('lorem', None, 'dolor', 'sit', None),
        ]

        for file_id, host_id, template_h, file_name, _ in cases:
            self.driver.translate_file(file_id, host_id, template_h,
                                       file_name)

        table = self.driver.state[cfgvalidator_driver.FILE]

        for _, _, _, _, expected in cases:
            if expected:
                self.assertIn(expected, table)

        expected_size = len(set([c[-1] for c in cases if c[-1]]))
        self.assertEqual(len(table), expected_size)

    def test_translate_template_ns(self):
        """Test the translation of namespace"""
        cases = [
            {
                'inputs': [
                    'lorem',
                    '',
                    {None: 'sit', 'amet': 'consectetur'}
                ],
                'expected': {
                    cfgvalidator_driver.TEMPLATE: (u'lorem', u''),
                    cfgvalidator_driver.NAMESPACE: (u'amet', u'consectetur'),
                    cfgvalidator_driver.TEMPLATE_NS: (u'lorem', u'amet'),
                }
            },
            {
                'inputs': [
                    '',
                    'ipsum',
                    {'dolor': 'sit', 'amet': ''}
                ],
                'expected': {
                    cfgvalidator_driver.TEMPLATE: None,
                    cfgvalidator_driver.NAMESPACE: None,
                    cfgvalidator_driver.TEMPLATE_NS: None,
                }
            },
            {
                'inputs': [
                    'lorem',
                    'ipsum',
                    {'dolor': 'sit'}
                ],
                'expected': {
                    cfgvalidator_driver.TEMPLATE: (u'lorem', u'ipsum'),
                    cfgvalidator_driver.NAMESPACE: (u'dolor', u'sit'),
                    cfgvalidator_driver.TEMPLATE_NS: (u'lorem', u'dolor'),
                }
            }
        ]

        for case in cases:
            self.driver.translate_template_namespace(*case['inputs'])

        for case in cases:
            for table_name, expected in six.iteritems(case['expected']):
                table = self.driver.state[table_name]

                if expected:
                    self.assertIn(expected, table)

        for table_name in [cfgvalidator_driver.TEMPLATE,
                           cfgvalidator_driver.NAMESPACE,
                           cfgvalidator_driver.TEMPLATE_NS]:
            expected_size = len(
                set([c['expected'][table_name] for c in cases
                     if c['expected'][table_name]]))
            table = self.driver.state[table_name]
            self.addDetail('table name', text_content(table_name))
            self.assertEqual(len(table), expected_size)

    def test_translate_option(self):
        """Unit tests for the translation of option definitions"""
        opt = cfg.StrOpt('host', required=True)
        opt.id_ = 'hash_opt'
        opt.ns_id = 'hash_ns'
        self.driver.translate_option(opt, "group")
        self.assertIsNotNone(self.driver.state['option'])
        self.assertIsNotNone(self.driver.state['option_info'])
        self.assertEqual(1, len(self.driver.state['option']))
        self.assertEqual(1, len(self.driver.state['option_info']))

    def test_translate_value(self):
        """Unit tests for translation of option values"""
        self.driver.translate_value("fid", 'optid1', 0)
        self.driver.translate_value("fid", 'optid2', [1, 2, 3])
        self.driver.translate_value("fid", 'optid3', {'a': 4, 'b': 5})
        self.assertEqual(6, len(self.driver.state['binding']))

    def test_translate_service(self):
        """Unit tests for translation of services"""
        self.driver.translate_service("hid", "svc", "vname")
        self.assertEqual(1, len(self.driver.state['service']))

    def test_process_template_hashes(self):
        """Test processing of template hash"""
        agent = self.agent_mock.return_value
        agent.get_template.return_value = {'namespaces': ['ns']}
        self.driver.process_template_hashes(['t1', 't2'], 'h')
        self.assertEqual(2, agent.get_template.call_count)
        self.assertEqual(1, agent.get_namespace.call_count)

    def test_translate_conf(self):
        """Test translation of conf"""
        self.driver.translate_conf(_fake_conf(), 'fid')
        state = self.driver.state
        self.assertEqual(2, len(state['option']))
        self.assertEqual(2, len(state['option_info']))
        self.assertEqual(2, len(state['binding']))

    @mock.patch('congress.cfg_validator.parsing.construct_conf_manager')
    @mock.patch('congress.cfg_validator.parsing.add_parsed_conf')
    def test_process_config(self, parsing_ccm, _):
        """Test complete processing of a conf"""
        parsing_ccm.return_value = _fake_conf()
        conf = {
            'template': 't',
            'service': 's',
            'version': 'v',
            'path': '/path/to/c',
            'data': {}
        }
        self.driver.known_templates['t'] = mock.MagicMock()
        self.driver.process_config('fhash', conf, 'h')
        state = self.driver.state
        self.assertEqual(1, len(state['service']))
        self.assertEqual(1, len(state['host']))
        self.assertEqual(1, len(state['template']))
        self.assertEqual(1, len(state['file']))

    @mock.patch('congress.cfg_validator.parsing.construct_conf_manager')
    @mock.patch('congress.cfg_validator.parsing.add_parsed_conf')
    def test_process_config_hashes(self, parsing_ccm, _):
        """Test processing of configuration hashes"""
        parsing_ccm.return_value = _fake_conf()
        conf = {
            'template': 't',
            'service': 's',
            'version': 'v',
            'path': '/path/to/c',
            'data': {}
        }
        self.agent_mock.return_value.get_config.return_value = conf
        self.driver.known_templates['t'] = mock.MagicMock()
        self.driver.process_config_hashes(['c'], 'h')
        state = self.driver.state
        self.assertEqual(1, len(state['service']))
        self.assertEqual(1, len(state['host']))
        self.assertEqual(1, len(state['template']))
        self.assertEqual(1, len(state['file']))

    def test_poll(self):
        """Test poll"""
        self.driver.poll()
        agt = self.agent_mock.return_value
        self.assertEqual(1, agt.publish_templates_hashes.call_count)
        self.assertEqual(1, agt.publish_configs_hashes.call_count)


class TestValidatorAgentClient(base_rpc.BaseTestRpcClient):
    """Unit tests for the RPC calls on the agent side"""
    def test_publish_config_hashes(self):
        "Test publish_config_hashes"
        rpcapi = cfgvalidator_driver.ValidatorAgentClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'publish_configs_hashes',
            rpc_method='cast', fanout=True
        )

    def test_publish_templates_hashes(self):
        "Test publish_templates_hashes"
        rpcapi = cfgvalidator_driver.ValidatorAgentClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'publish_templates_hashes',
            rpc_method='cast', fanout=True
        )

    def test_get_namespace(self):
        "test get_namespace"
        rpcapi = cfgvalidator_driver.ValidatorAgentClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'get_namespace',
            rpc_method='call', server="host",
            ns_hash='fake_hash'
        )

    # block calling thread
    def test_get_template(self):
        "test get_template"
        rpcapi = cfgvalidator_driver.ValidatorAgentClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'get_template',
            rpc_method='call', server="host",
            tpl_hash='fake_hash'
        )

    # block calling thread
    def test_get_config(self):
        "test get_config"
        rpcapi = cfgvalidator_driver.ValidatorAgentClient()
        self._test_rpc_api(
            rpcapi,
            None,
            'get_config',
            rpc_method='call', server="host",
            cfg_hash='fake_hash'
        )



# Copyright 2014 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import socket

from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
from oslo_middleware import cors

from congress import version

LOG = logging.getLogger(__name__)

core_opts = [
    cfg.HostAddressOpt('bind_host', default='0.0.0.0',
                       help="The host IP to bind to"),
    cfg.PortOpt('bind_port', default=1789,
                help="The port to bind to"),
    cfg.IntOpt('max_simultaneous_requests', default=1024,
               help="Thread pool size for eventlet."),
    cfg.BoolOpt('tcp_keepalive', default=False,
                help='Set this to true to enable TCP_KEEALIVE socket option '
                     'on connections received by the API server.'),
    cfg.IntOpt('tcp_keepidle',
               default=600,
               help='Sets the value of TCP_KEEPIDLE in seconds for each '
                    'server socket. Only applies if tcp_keepalive is '
                    'true. Not supported on OS X.'),
    cfg.IntOpt('api_workers', default=1,
               help='The number of worker processes to serve the congress '
                    'API application.'),
    cfg.StrOpt('api_paste_config', default='api-paste.ini',
               help=_('The API paste config file to use')),
    cfg.StrOpt('auth_strategy', default='keystone',
               help=_('The type of authentication to use')),
    cfg.ListOpt('drivers',
                default=[],
                deprecated_for_removal=True,
                deprecated_reason='automatically loads all configured drivers',
                help=_('List of driver class paths to import.')),
    cfg.ListOpt('disabled_drivers',
                default=[],
                help=_('List of driver names to be disabled. For example, '
                       'disabled_drivers=nova, plexxi')),
    cfg.ListOpt('custom_driver_endpoints',
                default=[],
                help=_("List of third party endpoints to be loaded seperated "
                       "by comma. For example custom_driver_endpoints = "
                       "'test=congress.datasources.test_driver:TestDriver',")),
    cfg.IntOpt('datasource_sync_period', default=60,
               help='The number of seconds to wait between synchronizing '
                    'datasource config from the database'),
    cfg.BoolOpt('enable_execute_action', default=True,
                help='Set the flag to False if you don\'t want Congress '
                     'to execute actions.'),
    cfg.BoolOpt('replicated_policy_engine', default=False,
                help='Set the flag to use congress with replicated policy '
                     'engines.'),
    cfg.StrOpt('policy_library_path', default='/etc/congress/library',
               help=_('The directory containing library policy files.')),
    cfg.StrOpt('encryption_key_path', default='/etc/congress/keys',
               help=_('The directory containing encryption keys.')),
]

# Register the configuration options
cfg.CONF.register_opts(core_opts)

dse_opts = [
    cfg.StrOpt('bus_id', default='bus',
               help='Unique ID of this DSE bus'),
    cfg.IntOpt('ping_timeout', default=5,
               help='RPC short timeout in seconds; used to ping destination'),
    cfg.IntOpt('long_timeout', default=120,
               help='RPC long timeout in seconds; used on potentially long '
                    'running requests such as datasource action and PE row '
                    'query'),
    cfg.IntOpt('time_to_resub', default=10,
               help='Time in seconds which a subscriber will wait for missing '
                    'update before attempting to resubscribe from publisher'),
    cfg.BoolOpt('execute_action_retry', default=False,
                help='Set the flag to True to make Congress retry execute '
                     'actions; may cause duplicate executions.'),
    cfg.IntOpt('execute_action_retry_timeout', default=600,
               help='The number of seconds to retry execute action before '
                    'giving up. Zero or negative value means never give up.'),
]

# Register dse opts
cfg.CONF.register_opts(dse_opts, group='dse')

logging.register_options(cfg.CONF)

_SQL_CONNECTION_DEFAULT = 'sqlite://'
# Update the default QueuePool parameters. These can be tweaked by the
# configuration variables - max_pool_size, max_overflow and pool_timeout
db_options.set_defaults(cfg.CONF,
                        connection=_SQL_CONNECTION_DEFAULT,
                        max_pool_size=10, max_overflow=20, pool_timeout=10)

# Command line options
cli_opts = [
    cfg.BoolOpt('datasources', default=False,
                help='Use this option to deploy the datasources.'),
    cfg.BoolOpt('api', default=False,
                help='Use this option to deploy API service'),
    cfg.BoolOpt('policy-engine', default=False,
                help='Use this option to deploy policy engine service.'),
    cfg.StrOpt('node-id', default=socket.gethostname(),
               help='A unique ID for this node.  Must be unique across all '
                    'nodes with the same bus_id.'),
    cfg.BoolOpt('delete-missing-driver-datasources', default=False,
                help='Use this option to delete datasources with missing '
                     'drivers from DB')
]
cfg.CONF.register_cli_opts(cli_opts)


def init(args, **kwargs):
    cfg.CONF(args=args, project='congress',
             version='%%(prog)s %s' % version.version_info.release_string(),
             **kwargs)


def setup_logging():
    """Sets up logging for the congress package."""
    logging.setup(cfg.CONF, 'congress')


def find_paste_config():
    config_path = cfg.CONF.find_file(cfg.CONF.api_paste_config)

    if not config_path:
        raise cfg.ConfigFilesNotFoundError(
            config_files=[cfg.CONF.api_paste_config])
    config_path = os.path.abspath(config_path)
    LOG.info(("Config paste file: %s"), config_path)
    return config_path


def set_config_defaults():
    """This method updates all configuration default values."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-OpenStack-Request-ID',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id'],
        expose_headers=['X-Auth-Token',
                        'X-OpenStack-Request-ID',
                        'X-Subject-Token',
                        'X-Service-Token'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
from oslo_config import cfg
from six.moves import reduce

from congress.api import webservice
from congress.datasources import nova_driver
from congress import exception
from congress.tests.api import base as api_base
from congress.tests import base
from congress.tests import helper


class TestDatasourceModel(base.SqlTestCase):
    def setUp(self):
        super(TestDatasourceModel, self).setUp()
        cfg.CONF.set_override(
            'custom_driver_endpoints',
            'test=congress.tests.test_custom_driver:TestCustomDriver')
        services = api_base.setup_config(with_fake_datasource=False)
        self.datasource_model = services['api']['api-datasource']
        self.data = services['data']
        self.node = services['node']
        self.engine = services['engine']
        self.ds_manager = services['ds_manager']
        self.datasource = self._get_datasource_request()
        self.ds_manager.add_datasource(self.datasource)

    def tearDown(self):
        super(TestDatasourceModel, self).tearDown()
        self.node.stop()

    def _get_datasource_request(self):
        # leave ID out--generated during creation
        return {'name': 'datasource1',
                'driver': 'fake_datasource',
                'description': 'hello world!',
                'enabled': True,
                'type': None,
                'config': {'auth_url': 'foo',
                           'username': 'armax',
                           'password': '<hidden>',
                           'tenant_name': 'armax'}}

    def test_get_items(self):
        dinfo = self.datasource_model.get_items(None)['results']
        self.assertEqual(1, len(dinfo))
        datasource2 = self._get_datasource_request()
        datasource2['name'] = 'datasource2'
        self.ds_manager.add_datasource(datasource2)
        dinfo = self.datasource_model.get_items(None)['results']
        self.assertEqual(2, len(dinfo))
        del dinfo[0]['id']
        self.assertEqual(self.datasource, dinfo[0])

    def test_add_item(self):
        datasource3 = self._get_datasource_request()
        datasource3['name'] = 'datasource_test_3'
        self.datasource_model.add_item(datasource3, {})
        ds_obj = self.node.service_object('datasource_test_3')
        obj = self.engine.policy_object('datasource_test_3')
        self.assertIsNotNone(obj.schema)
        self.assertEqual('datasource_test_3', obj.name)
        self.assertIsNotNone(ds_obj)

    def test_add_datasource_with_custom_driver(self):
        datasource4 = self._get_datasource_request()
        datasource4['name'] = 'datasource_test_4'
        datasource4['driver'] = 'test'
        self.datasource_model.add_item(datasource4, {})
        ds_obj = self.node.service_object('datasource_test_4')
        obj = self.engine.policy_object('datasource_test_4')
        self.assertIsNotNone(obj.schema)
        self.assertEqual('datasource_test_4', obj.name)
        self.assertIsNotNone(ds_obj)

    def test_add_item_duplicate(self):
        self.assertRaises(webservice.DataModelException,
                          self.datasource_model.add_item,
                          self.datasource, {})

    def test_add_item_invalid_tablename(self):
        datasource = self._get_datasource_request()
        datasource['name'] = "invalid-table-name"
        self.assertRaises(webservice.DataModelException,
                          self.datasource_model.add_item,
                          datasource, {})

    def test_delete_item(self):
        datasource = self._get_datasource_request()
        datasource['name'] = 'test_datasource'
        d_id, dinfo = self.datasource_model.add_item(datasource, {})
        self.assertTrue(self.engine.assert_policy_exists('test_datasource'))
        context = {'ds_id': d_id}
        self.datasource_model.delete_item(None, {}, context=context)
        ds_obj = self.node.service_object('test_datasource')
        self.assertIsNone(ds_obj)
        self.assertRaises(exception.PolicyRuntimeException,
                          self.engine.assert_policy_exists, 'test_datasource')
        self.assertRaises(exception.DatasourceNotFound,
                          self.node.get_datasource, d_id)

    def test_delete_item_invalid_datasource(self):
        context = {'ds_id': 'fake'}
        self.assertRaises(webservice.DataModelException,
                          self.datasource_model.delete_item,
                          None, {}, context=context)

    def test_datasource_api_model_execute(self):
        def _execute_api(client, action, action_args):
            positional_args = action_args.get('positional', [])
            named_args = action_args.get('named', {})
            method = reduce(getattr, action.split('.'), client)
            method(*positional_args, **named_args)

        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def _get_testkey(self):
                return self.testkey

            def disconnect(self, arg1, arg2, arg3):
                self.testkey = "arg1=%s arg2=%s arg3=%s" % (arg1, arg2, arg3)

            def disconnect_all(self):
                self.testkey = "action_has_no_args"

        nova_client = NovaClient("testing")
        args = helper.datasource_openstack_args()
        nova = nova_driver.NovaDriver('nova', args=args)
        nova.nova_client = nova_client
        nova.update_from_datasource = mock.MagicMock()
        nova._execute_api = _execute_api
        self.node.register_service(nova)

        execute_action = self.datasource_model.execute_action

        # Positive test: valid body args, ds_id
        context = {'ds_id': 'nova'}
        body = {'name': 'disconnect',
                'args': {'positional': ['value1', 'value2'],
                         'named': {'arg3': 'value3'}}}
        request = helper.FakeRequest(body)
        result = execute_action({}, context, request)
        self.assertEqual(result, {})
        expected_result = "arg1=value1 arg2=value2 arg3=value3"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)

        # Positive test: no body args
        context = {'ds_id': 'nova'}
        body = {'name': 'disconnect_all'}
        request = helper.FakeRequest(body)
        result = execute_action({}, context, request)
        self.assertEqual(result, {})
        expected_result = "action_has_no_args"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)

        # Negative test: invalid ds_id
        context = {'ds_id': 'unknown_ds'}
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, request)

        # Negative test: no ds_id
        context = {}
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, request)

        # Negative test: empty body
        context = {'ds_id': 'nova'}
        bad_request = helper.FakeRequest({})
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, bad_request)

        # Negative test: no body name/action
        context = {'ds_id': 'nova'}
        body = {'args': {'positional': ['value1', 'value2'],
                         'named': {'arg3': 'value3'}}}
        bad_request = helper.FakeRequest(body)
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, bad_request)

        # Positive test with retry: no body args
        cfg.CONF.dse.execute_action_retry = True
        context = {'ds_id': 'nova'}
        body = {'name': 'disconnect_all'}
        request = helper.FakeRequest(body)
        result = execute_action({}, context, request)
        self.assertEqual(result, {})
        expected_result = "action_has_no_args"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_utils import uuidutils

from congress.db import db_ds_table_data
from congress.tests import base


class TestDbDsTableData(base.SqlTestCase):

    def test_store_ds_table_data(self):
        ds_id = uuidutils.generate_uuid()

        # store new data
        source = db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table1',
            tabledata=set([('a', 0)]))
        self.assertEqual(ds_id, source.ds_id)
        self.assertEqual('table1', source.tablename)
        self.assertEqual('[["a", 0]]', source.tabledata)

        # update exsting data
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table1',
            tabledata=set([('a', 0), ('new', 1)]))
        data = db_ds_table_data.get_ds_table_data(ds_id, 'table1')
        self.assertEqual(set([('a', 0), ('new', 1)]), data)

    def test_delete_ds_table_data(self):
        ds_id = uuidutils.generate_uuid()
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table1',
            tabledata=set([('a', 0), ('b', 1)]))
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table2',
            tabledata=set([('a', 0), ('b', 2)]))
        self.assertTrue(db_ds_table_data.delete_ds_table_data(ds_id, 'table1'))
        self.assertIsNone(db_ds_table_data.get_ds_table_data(ds_id, 'table1'))
        self.assertEqual(set([('a', 0), ('b', 2)]),
                         db_ds_table_data.get_ds_table_data(ds_id, 'table2'))

    def test_delete_ds_table_data_by_ds(self):
        ds_id = uuidutils.generate_uuid()
        ds2_id = uuidutils.generate_uuid()
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table1',
            tabledata=set([('a', 0), ('b', 1)]))
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table2',
            tabledata=set([('a', 0), ('b', 2)]))
        db_ds_table_data.store_ds_table_data(
            ds_id=ds2_id,
            tablename='table3',
            tabledata=set([('a', 0), ('b', 3)]))
        self.assertTrue(db_ds_table_data.delete_ds_table_data(ds_id))
        self.assertIsNone(db_ds_table_data.get_ds_table_data(ds_id, 'table1'))
        self.assertIsNone(db_ds_table_data.get_ds_table_data(ds_id, 'table2'))
        self.assertEqual(set([('a', 0), ('b', 3)]),
                         db_ds_table_data.get_ds_table_data(ds2_id, 'table3'))

    def test_delete_non_existing_ds_table_data(self):
        self.assertFalse(db_ds_table_data.delete_ds_table_data('none', 'none'))

    def test_get_ds_table_data(self):
        ds_id = uuidutils.generate_uuid()
        ds2_id = uuidutils.generate_uuid()
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table1',
            tabledata=set([('a', 0), ('b', 1)]))
        db_ds_table_data.store_ds_table_data(
            ds_id=ds_id,
            tablename='table2',
            tabledata=set([('a', 0), ('b', 2)]))
        db_ds_table_data.store_ds_table_data(
            ds_id=ds2_id,
            tablename='table3',
            tabledata=set([('a', 0), ('b', 3)]))
        data = db_ds_table_data.get_ds_table_data(ds_id, 'table1')
        self.assertEqual(set([('a', 0), ('b', 1)]), data)

        data = db_ds_table_data.get_ds_table_data(ds_id)
        self.assertEqual(2, len(data))
        self.assertIn(
            {'tablename': 'table1',
             'tabledata': set([('a', 0), ('b', 1)])},
            data)
        self.assertIn(
            {'tablename': 'table2',
             'tabledata': set([('a', 0), ('b', 2)])},
            data)



# Copyright (c) 2017 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Encryption module for handling passwords in Congress."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import io
import os

from cryptography import fernet
from cryptography.fernet import Fernet
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

__key = None
__fernet = None


def _get_key_file_path():
    return os.path.join(cfg.CONF.encryption_key_path, 'aes_key')


def key_file_exists():
    return os.path.isfile(_get_key_file_path())


def read_key_from_file():
    with io.open(_get_key_file_path(), 'r', encoding='ascii') as key_file:
        key = str(key_file.read()).encode('ascii')
    return key


def create_new_key_file():
    LOG.debug("Generate new key file")
    dir_path = os.path.dirname(_get_key_file_path())
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path, mode=0o700)  # important: restrictive permissions
    key = Fernet.generate_key()
    # first create file with restrictive permissions, then write key
    # two separate file opens because each version supports
    # permissions and encoding respectively, but neither supports both.
    with os.fdopen(os.open(_get_key_file_path(), os.O_CREAT | os.O_WRONLY,
                           0o600), 'w'):
        pass
    with io.open(_get_key_file_path(), 'w', encoding='ascii') as key_file:
        key_file.write(key.decode('ascii'))
    return key


def initialize_key():
    '''initialize key.'''
    global __key
    global __fernet
    if key_file_exists():
        __key = read_key_from_file()
    else:
        __key = create_new_key_file()

    __fernet = Fernet(__key)


def initialize_if_needed():
    '''initialize key if not already initialized.'''
    global __fernet
    if not __fernet:
        initialize_key()


def encrypt(string):
    initialize_if_needed()
    return __fernet.encrypt(string.encode('utf-8')).decode('utf-8')


class InvalidToken(fernet.InvalidToken):
    pass


def decrypt(string):
    initialize_if_needed()
    try:
        return __fernet.decrypt(string.encode('utf-8')).decode('utf-8')
    except fernet.InvalidToken as exc:
        raise InvalidToken(exc)



#!/usr/bin/python
# Copyright (c) 2017 NTT All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This script outputs pre definded policies and rules in a config file.
# The config file content should look like following:
# {
#   "policies":
#     [
#       {
#         "name": "sample_policy",
#         "rules": [
#                    {
#                      "rule": "p(x):- q(x)",
#                      "name": "rule1"
#                    },
#                    {
#                      "rule": "q(1)"
#                    },
#                    {
#                      "rule": "q('sample-row')"
#                    },
#                    {
#                      "rule": "server_ids(x):- nova:servers(id=x)"
#                    }
#         ]
#       }
#    ]
# }
#
# Config file option:
#    - "name" key in rule object is option
#
# sample config file is located at
# /path/to/congress/scripts/preload-polices/policy-rules.json.sample


import argparse
import json
import sys

OPENSTACK_COMMAND = 'openstack congress'
POLICY_CREATE = 'policy create'
RULE_CREATE = 'policy rule create'
NAME_OPTION = '--name %s'


def load_policies(policy_file):
    with open(args.policy_file, 'r') as f:
        data = f.read()
        policies = json.loads(data)
    return policies


def main(args):
    defined_policies = load_policies(args.policy_file)

    for p in defined_policies['policies']:
        # create defined policy
        sys.stdout.write(' '.join([OPENSTACK_COMMAND, POLICY_CREATE,
                                   p['name'], '\n']))
        # create defined rules
        for r in p['rules']:
            cmd_string = [OPENSTACK_COMMAND, RULE_CREATE]
            if r.get('name'):
                cmd_string.append(NAME_OPTION % r['name'])
            cmd_string.extend([p['name'], '"%s"' % r['rule'], '\n'])
            sys.stdout.write(' '.join(cmd_string))


parser = argparse.ArgumentParser(description='Output pre-defined policy in '
                                             'openstack command style.')
parser.add_argument('policy_file', type=str,
                    help='Path to pre defined policies and rules.')


if __name__ == '__main__':
    args = parser.parse_args()
    main(args)



#    Copyright 2018 Orange
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""A Mock of Z3"""

Z3_OP_OR = 262
Z3_OP_AND = 261
Z3_OP_EQ = 258
Z3_OP_FALSE = 257
Z3_OP_TRUE = 256
Z3_OP_NOT = 265
Z3_OP_IMPLIES = 266

Z3_OP_BNUM = 1024
Z3_OP_UNINTERPRETED = 2354


class BoolSort(object):

    def __repr__(self):
        return 'bool'


class BitVecSort(object):
    def __init__(self, size):
        self._size = size

    def size(self):
        return self._size

    def __repr__(self):
        return 'BV({})'.format(self._size)


class FuncDeclRef(object):
    def __init__(self, name, kind, typs=None):
        self._name = name
        self._kind = kind
        self._typs = [] if typs is None else typs

    def name(self):
        return self._name

    def kind(self):
        return self._kind

    def __eq__(self, x):
        return (isinstance(x, self.__class__) and x._name == self._name and
                x._kind == self._kind)

    def __call__(self, *args):
        return ExprRef(self, args)

    def __repr__(self):
        return (self._name if self._typs == []
                else self._name + repr(self._typs))


class ExprRef(object):
    def __init__(self, decl, args):
        self._decl = decl
        self.args = args

    def decl(self):
        return self._decl

    def children(self):
        return self.args

    def __eq__(self, v):
        return ExprRef(FuncDeclRef('=', Z3_OP_EQ), [self, v])

    def __repr__(self):
        return repr(self._decl) + repr(self.args)


class BitVecRef(ExprRef):
    def __init__(self, decl, val, sort):
        super(BitVecRef, self).__init__(decl, [])
        self.val = val
        self.sort = sort

    def as_long(self):
        return self.val

    def size(self):
        return self.sort.size()

    def __repr__(self):
        return "bv({})".format(self.val)


def BitVecVal(v, s):
    return BitVecRef(FuncDeclRef('bv', Z3_OP_BNUM), v, s)


class BoolRef(ExprRef):
    pass


def BoolVal(val):
    return BoolRef(FuncDeclRef('true' if val else 'false',
                               Z3_OP_TRUE if val else Z3_OP_FALSE),
                   [])


def And(*args):
    return BoolRef(FuncDeclRef('and', Z3_OP_AND), args)


def Or(*args):
    return BoolRef(FuncDeclRef('or', Z3_OP_OR), args)


def Not(arg):
    return BoolRef(FuncDeclRef('not', Z3_OP_NOT), [arg])


def Eq(arg1, arg2):
    return BoolRef(FuncDeclRef('=', Z3_OP_EQ), [arg1, arg2])


def Implies(a, b):
    return BoolRef(FuncDeclRef("=>", Z3_OP_IMPLIES), [a, b])


def ForAll(l, t):
    return BoolRef(FuncDeclRef("forall", 0), l + [t])


def Exists(l, t):
    return BoolRef(FuncDeclRef("exists", 0), l + [t])


def Const(x, sort):
    return ExprRef(FuncDeclRef(x, Z3_OP_UNINTERPRETED, typs=[sort]), [])


def Function(f, *args):
    return FuncDeclRef(f, Z3_OP_UNINTERPRETED, typs=args)


class Fixedpoint(object):

    def __init__(self):
        self._relations = []
        self._rules = []
        self._options = {}

    def set(self, **kwargs):
        self._options = kwargs

    def register_relation(self, r):
        self._relations.append(r)

    def fact(self, f):
        self._rules.append(f)

    def rule(self, r):
        self._rules.append(r)

    def query(self, q):
        pass

    @staticmethod
    def get_answer():
        return BoolVal(True)

    def get_rules(self):
        return self._rules
